import datetime
import importlib
import numcodecs
import os
import pathlib
import time
from typing import Dict, Union

import numpy as np
import xarray as xr

import toolviper.utils.logger as logger
from casacore import tables
from xradio.measurement_set._utils._msv2.msv4_sub_xdss import (
    create_pointing_xds,
    create_system_calibration_xds,
    create_weather_xds,
)
from .msv4_info_dicts import create_info_dicts
from xradio.measurement_set._utils._msv2.create_antenna_xds import (
    create_antenna_xds,
    create_gain_curve_xds,
    create_phase_calibration_xds,
)
from xradio.measurement_set._utils._msv2.create_field_and_source_xds import (
    create_field_and_source_xds,
)
from xradio._utils.schema import column_description_casacore_to_msv4_measure
from .msv2_to_msv4_meta import (
    create_attribute_metadata,
    col_to_data_variable_names,
    col_dims,
)

from .._zarr.encoding import add_encoding
from .subtables import subt_rename_ids
from ._tables.table_query import open_table_ro, open_query
from ._tables.read import (
    convert_casacore_time,
    extract_table_attributes,
    read_col_conversion,
    load_generic_table,
)
from ._tables.read_main_table import get_baselines, get_baseline_indices, get_utimes_tol
from .._utils.stokes_types import stokes_types
from xradio._utils.list_and_array import check_if_consistent, unique_1d, to_list


def parse_chunksize(
    chunksize: Union[Dict, float, None], xds_type: str, xds: xr.Dataset
) -> Dict[str, int]:
    """
    Parameters
    ----------
    chunksize : Union[Dict, float, None]
        Desired maximum size of the chunks, either as a dict of per-dimension sizes or as
        an amount of memory
    xds_type : str
        whether to use chunking logic for main or pointing datasets
    xds : xr.Dataset
        dataset to calculate best chunking

    Returns
    -------
    Dict[str, int]
        dictionary of chunk sizes (as dim->size)
    """
    if isinstance(chunksize, dict):
        check_chunksize(chunksize, xds_type)
    elif isinstance(chunksize, float):
        chunksize = mem_chunksize_to_dict(chunksize, xds_type, xds)
    elif chunksize is not None:
        raise ValueError(
            f"Chunk size expected as a dict or a float, got: "
            f" {chunksize} (of type {type(chunksize)}"
        )

    return chunksize


def check_chunksize(chunksize: dict, xds_type: str) -> None:
    """
    Rudimentary check of the chunksize parameters to catch obvious errors early before
    more work is done.
    """
    # perphaps start using some TypeDict or/and validator like pydantic?
    if xds_type == "main":
        allowed_dims = [
            "time",
            "baseline_id",
            "antenna_name",
            "frequency",
            "polarization",
        ]
    elif xds_type == "pointing":
        allowed_dims = ["time", "antenna"]

    msg = ""
    for dim in chunksize.keys():
        if dim not in allowed_dims:
            msg += f"dimension {dim} not allowed in {xds_type} dataset:\n"
    if msg:
        raise ValueError(f"Wrong keys found in chunksize: {msg}")


def mem_chunksize_to_dict(
    chunksize: float, xds_type: str, xds: xr.Dataset
) -> Dict[str, int]:
    """
    Given a desired 'chunksize' as amount of memory in GB, calculate best chunk sizes
    for every dimension of an xds.

    Parameters
    ----------
    chunksize : float
        Desired maximum size of the chunks
    xds_type : str
        whether to use chunking logic for main or pointing datasets
    xds : xr.Dataset
        dataset to auto-calculate chunking of its dimensions

    Returns
    -------
    Dict[str, int]
        dictionary of chunk sizes (as dim->size)
    """

    if xds_type == "pointing":
        sizes = mem_chunksize_to_dict_pointing(chunksize, xds)
    elif xds_type == "main":
        sizes = mem_chunksize_to_dict_main(chunksize, xds)
    else:
        raise RuntimeError(f"Unexpected type: {xds_type=}")

    return sizes


GiBYTES_TO_BYTES = 1024 * 1024 * 1024


def mem_chunksize_to_dict_main(chunksize: float, xds: xr.Dataset) -> Dict[str, int]:
    """
    Checks the assumption that all polarizations can be held in memory, at least for one
    data point (one time, one freq, one channel).

    It presently relies on the logic of mem_chunksize_to_dict_main_balanced() to find a
    balanced list of dimension sizes for the chunks

    Assumes these relevant dims: (time, antenna_name/baseline_id, frequency,
    polarization).
    """

    sizeof_vis = itemsize_spec(xds)
    size_all_pols = sizeof_vis * xds.sizes["polarization"]
    if size_all_pols / GiBYTES_TO_BYTES > chunksize:
        raise RuntimeError(
            "Cannot calculate chunk sizes when memory bound ({chunksize}) does not even allow all polarizations in one chunk"
        )

    baseline_or_antenna_name = find_baseline_or_antenna_var(xds)
    total_size = calc_used_gb(xds.sizes, baseline_or_antenna_name, sizeof_vis)

    ratio = chunksize / total_size
    chunked_dims = ["time", baseline_or_antenna_name, "frequency", "polarization"]
    if ratio >= 1:
        result = {dim: xds.sizes[dim] for dim in chunked_dims}
        logger.debug(
            f"{chunksize=} GiB is enough to fully hold {total_size=} GiB (for {xds.sizes=}) in memory in one chunk"
        )
    else:
        xds_dim_sizes = {k: xds.sizes[k] for k in chunked_dims}
        result = mem_chunksize_to_dict_main_balanced(
            chunksize, xds_dim_sizes, baseline_or_antenna_name, sizeof_vis
        )

    return result


def mem_chunksize_to_dict_main_balanced(
    chunksize: float,
    xds_dim_sizes: dict,
    baseline_or_antenna_name: str,
    sizeof_vis: int,
) -> Dict[str, int]:
    """
    Assumes the ratio is <1 and all pols can fit in memory (from
    mem_chunksize_to_dict_main()).

    What is kept balanced is the fraction of the total size of every dimension included in a
    chunk. For example, time: 10, baseline: 100, freq: 1000, if we can afford about 33% in
    one chunk, the chunksize will be ~ time: 3, baseline: 33, freq: 333.
    The polarization axis is excluded from the calculations.
    Because this can leave a leftover (below or above the desired chunksize limit) and
    adjustment is done to get the final memory use below but as close as possible to
    'chunksize'. This adjustment alters the balance.

    Parameters
    ----------
    chunksize : float
        Desired maximum size of the chunks
    xds_dim_sizes : dict
        Dataset dimension sizes as dim_name->size
    sizeof_vis : int
        Size in bytes of a data point (one visibility / spectrum value)

    Returns
    -------
    Dict[str, int]
        dictionary of chunk sizes (as dim->size)
    """

    dim_names = [name for name in xds_dim_sizes.keys()]
    dim_sizes = [size for size in xds_dim_sizes.values()]
    # Fix fourth dim (polarization) to all (not free to auto-calculate)
    free_dims_mask = np.array([True, True, True, False])

    total_size = np.prod(dim_sizes) * sizeof_vis / GiBYTES_TO_BYTES
    ratio = chunksize / total_size

    dim_chunksizes = np.array(dim_sizes, dtype="int64")
    factor = ratio ** (1 / np.sum(free_dims_mask))
    dim_chunksizes[free_dims_mask] = np.maximum(
        dim_chunksizes[free_dims_mask] * factor, 1
    )
    used = np.prod(dim_chunksizes) * sizeof_vis / GiBYTES_TO_BYTES

    logger.debug(
        f"Auto-calculating main chunk sizes. First order approximation {dim_chunksizes=}, used total: {used} GiB (with {chunksize=} GiB)"
    )

    # Iterate through the dims, starting from the dims with lower chunk size
    #  (=bigger impact of a +1)
    # Note the use of np.floor, this iteration can either increase or decrease sizes,
    #  if increasing sizes we want to keep mem use below the upper limit, floor(2.3) = +2
    #  if decreasing sizes we want to take mem use below the upper limit, floor(-2.3) = -3
    indices = np.argsort(dim_chunksizes[free_dims_mask])
    for idx in indices:
        left = chunksize - used
        other_dims_mask = np.ones(free_dims_mask.shape, dtype=bool)
        other_dims_mask[idx] = False
        delta = np.divide(
            left,
            np.prod(dim_chunksizes[other_dims_mask]) * sizeof_vis / GiBYTES_TO_BYTES,
        )
        int_delta = np.floor(delta)
        if abs(int_delta) > 0 and int_delta + dim_chunksizes[idx] > 0:
            dim_chunksizes[idx] += int_delta
        used = np.prod(dim_chunksizes) * sizeof_vis / GiBYTES_TO_BYTES

    chunked_dim_names = ["time", baseline_or_antenna_name, "frequency", "polarization"]
    dim_chunksizes_int = [int(v) for v in dim_chunksizes]
    result = dict(zip(chunked_dim_names, dim_chunksizes_int))

    logger.debug(
        f"Auto-calculated main chunk sizes with {chunksize=}, {total_size=} GiB (for {dim_sizes=}): {result=} which uses {used} GiB."
    )

    return result


def mem_chunksize_to_dict_pointing(chunksize: float, xds: xr.Dataset) -> Dict[str, int]:
    """
    Equivalent to mem_chunksize_to_dict_main adapted to pointing xdss.
    Assumes these relevant dims: (time, antenna, direction).
    """

    if not xds.sizes:
        return {}

    sizeof_pointing = itemsize_pointing_spec(xds)
    chunked_dim_names = [name for name in xds.sizes.keys()]
    dim_sizes = [size for size in xds.sizes.values()]
    total_size = np.prod(dim_sizes) * sizeof_pointing / GiBYTES_TO_BYTES

    # Fix third dim (direction) to all
    free_dims_mask = np.array([True, True, False])

    ratio = chunksize / total_size
    if ratio >= 1:
        logger.debug(
            f"Pointing chunsize: {chunksize=} GiB is enough to fully hold {total_size=} GiB (for {xds.sizes=}) in memory in one chunk"
        )
        dim_chunksizes = dim_sizes
    else:
        # balanced
        dim_chunksizes = np.array(dim_sizes, dtype="int")
        factor = ratio ** (1 / np.sum(free_dims_mask))
        dim_chunksizes[free_dims_mask] = np.maximum(
            dim_chunksizes[free_dims_mask] * factor, 1
        )
        used = np.prod(dim_chunksizes) * sizeof_pointing / GiBYTES_TO_BYTES

        logger.debug(
            f"Auto-calculating pointing chunk sizes. First order approximation: {dim_chunksizes=}, used total: {used=} GiB (with {chunksize=} GiB"
        )

        indices = np.argsort(dim_chunksizes[free_dims_mask])
        # refine dim_chunksizes
        for idx in indices:
            left = chunksize - used
            other_dims_mask = np.ones(free_dims_mask.shape, dtype=bool)
            other_dims_mask[idx] = False
            delta = np.divide(
                left,
                np.prod(dim_chunksizes[other_dims_mask])
                * sizeof_pointing
                / GiBYTES_TO_BYTES,
            )
            int_delta = np.floor(delta)
            if abs(int_delta) > 0 and int_delta + dim_chunksizes[idx] > 0:
                dim_chunksizes[idx] += int_delta

            used = np.prod(dim_chunksizes) * sizeof_pointing / GiBYTES_TO_BYTES

    dim_chunksizes_int = [int(v) for v in dim_chunksizes]
    result = dict(zip(chunked_dim_names, dim_chunksizes_int))

    if ratio < 1:
        logger.debug(
            f"Auto-calculated pointing chunk sizes with {chunksize=}, {total_size=} GiB (for {xds.sizes=}): {result=} which uses {used} GiB."
        )

    return result


def find_baseline_or_antenna_var(xds: xr.Dataset) -> str:
    if "baseline_id" in xds.coords:
        baseline_or_antenna_name = "baseline_id"
    elif "antenna_name" in xds.coords:
        baseline_or_antenna_name = "antenna_name"

    return baseline_or_antenna_name


def itemsize_spec(xds: xr.Dataset) -> int:
    """
    Size in bytes of one visibility (or spectrum) value.
    """
    names = ["SPECTRUM", "VISIBILITY"]
    itemsize = 8
    for var in names:
        if var in xds.data_vars:
            var_name = var
            itemsize = np.dtype(xds.data_vars[var_name].dtype).itemsize
            break

    return itemsize


def itemsize_pointing_spec(xds: xr.Dataset) -> int:
    """
    Size in bytes of one pointing (or spectrum) value.
    """
    pnames = ["BEAM_POINTING"]
    itemsize = 8
    for var in pnames:
        if var in xds.data_vars:
            var_name = var
            itemsize = np.dtype(xds.data_vars[var_name].dtype).itemsize
            break

    return itemsize


def calc_used_gb(
    chunksizes: dict, baseline_or_antenna_name: str, sizeof_vis: int
) -> float:
    return (
        chunksizes["time"]
        * chunksizes[baseline_or_antenna_name]
        * chunksizes["frequency"]
        * chunksizes["polarization"]
        * sizeof_vis
        / GiBYTES_TO_BYTES
    )


# TODO: if the didxs are not used in read_col_conversion, remove didxs from here (and convert_and_write_partition)
def calc_indx_for_row_split(tb_tool, taql_where):
    baselines = get_baselines(tb_tool)
    col_names = tb_tool.colnames()
    cshapes = [
        np.array(tb_tool.getcell(col, 0)).shape
        for col in col_names
        if tb_tool.iscelldefined(col, 0)
    ]

    freq_cnt, pol_cnt = [(cc[0], cc[1]) for cc in cshapes if len(cc) == 2][0]
    utimes, tol = get_utimes_tol(tb_tool, taql_where)

    tidxs = np.searchsorted(utimes, tb_tool.getcol("TIME"))

    ts_ant1, ts_ant2 = (
        tb_tool.getcol("ANTENNA1"),
        tb_tool.getcol("ANTENNA2"),
    )

    ts_bases = np.column_stack((ts_ant1, ts_ant2))
    bidxs = get_baseline_indices(baselines, ts_bases)

    # some antenna 2"s will be out of bounds for this chunk, store rows that are in bounds

    didxs = np.where((bidxs >= 0) & (bidxs < len(baselines)))[0]

    baseline_ant1_id = baselines[:, 0]
    baseline_ant2_id = baselines[:, 1]

    return (
        tidxs,
        bidxs,
        didxs,
        baseline_ant1_id,
        baseline_ant2_id,
        convert_casacore_time(utimes, False),
    )


def create_coordinates(
    xds, in_file, ddi, utime, interval, baseline_ant1_id, baseline_ant2_id, scan_id
):
    coords = {
        "time": utime,
        "baseline_antenna1_id": ("baseline_id", baseline_ant1_id),
        "baseline_antenna2_id": ("baseline_id", baseline_ant2_id),
        "baseline_id": np.arange(len(baseline_ant1_id)),
        "scan_number": ("time", scan_id),
        "uvw_label": ["u", "v", "w"],
    }

    ddi_xds = load_generic_table(in_file, "DATA_DESCRIPTION").sel(row=ddi)
    pol_setup_id = ddi_xds.POLARIZATION_ID.values
    spectral_window_id = int(ddi_xds.SPECTRAL_WINDOW_ID.values)

    spectral_window_xds = load_generic_table(
        in_file,
        "SPECTRAL_WINDOW",
        rename_ids=subt_rename_ids["SPECTRAL_WINDOW"],
    ).sel(spectral_window_id=spectral_window_id)
    coords["frequency"] = spectral_window_xds["CHAN_FREQ"].data[
        ~(np.isnan(spectral_window_xds["CHAN_FREQ"].data))
    ]

    pol_xds = load_generic_table(
        in_file,
        "POLARIZATION",
        rename_ids=subt_rename_ids["POLARIZATION"],
    )
    num_corr = int(pol_xds["NUM_CORR"][pol_setup_id].values)
    coords["polarization"] = np.vectorize(stokes_types.get)(
        pol_xds["CORR_TYPE"][pol_setup_id, :num_corr].values
    )

    xds = xds.assign_coords(coords)

    ###### Create Frequency Coordinate ######
    freq_column_description = spectral_window_xds.attrs["other"]["msv2"]["ctds_attrs"][
        "column_descriptions"
    ]

    msv4_measure = column_description_casacore_to_msv4_measure(
        freq_column_description["CHAN_FREQ"],
        ref_code=spectral_window_xds["MEAS_FREQ_REF"].data,
    )
    xds.frequency.attrs.update(msv4_measure)

    spw_name = spectral_window_xds.NAME.values.item()
    if (spw_name is None) or (spw_name == "none") or (spw_name == ""):
        spw_name = "spw_" + str(spectral_window_id)
    else:
        # spw_name = spectral_window_xds.NAME.values.item()
        spw_name = spw_name + "_" + str(spectral_window_id)

    xds.frequency.attrs["spectral_window_name"] = spw_name
    msv4_measure = column_description_casacore_to_msv4_measure(
        freq_column_description["REF_FREQUENCY"],
        ref_code=spectral_window_xds["MEAS_FREQ_REF"].data,
    )
    xds.frequency.attrs["reference_frequency"] = {
        "dims": [],
        "data": float(spectral_window_xds.REF_FREQUENCY.values),
        "attrs": msv4_measure,
    }
    xds.frequency.attrs["spectral_window_id"] = spectral_window_id

    # Add if doppler table is present
    # xds.frequency.attrs["doppler_velocity"] =
    # xds.frequency.attrs["doppler_type"] =

    unique_chan_width = unique_1d(
        spectral_window_xds["CHAN_WIDTH"].data[
            np.logical_not(np.isnan(spectral_window_xds["CHAN_WIDTH"].data))
        ]
    )
    # assert len(unique_chan_width) == 1, "Channel width varies for spectral_window."
    # xds.frequency.attrs["channel_width"] = spectral_window_xds.chan_width.data[
    #    ~(np.isnan(spectral_window_xds.chan_width.data))
    # ]  # unique_chan_width[0]
    msv4_measure = column_description_casacore_to_msv4_measure(
        freq_column_description["CHAN_WIDTH"],
        ref_code=spectral_window_xds["MEAS_FREQ_REF"].data,
    )
    if not msv4_measure:
        msv4_measure["type"] = "quantity"
        msv4_measure["units"] = ["Hz"]
    xds.frequency.attrs["channel_width"] = {
        "dims": [],
        "data": np.abs(unique_chan_width[0]),
        "attrs": msv4_measure,
    }

    ###### Create Time Coordinate ######
    main_table_attrs = extract_table_attributes(in_file)
    main_column_descriptions = main_table_attrs["column_descriptions"]
    msv4_measure = column_description_casacore_to_msv4_measure(
        main_column_descriptions["TIME"]
    )
    xds.time.attrs.update(msv4_measure)

    msv4_measure = column_description_casacore_to_msv4_measure(
        main_column_descriptions["INTERVAL"]
    )
    if not msv4_measure:
        msv4_measure["type"] = "quantity"
        msv4_measure["units"] = ["s"]
    xds.time.attrs["integration_time"] = {
        "dims": [],
        "data": interval,
        "attrs": msv4_measure,
    }
    xds.time.attrs["effective_integration_time"] = "EFFECTIVE_INTEGRATION_TIME"
    return xds


def find_min_max_times(tb_tool: tables.table, taql_where: str) -> tuple:
    """
    Find the min/max times in an MSv4, for constraining pointing.

    To avoid numerical comparison issues (leaving out some times at the edges),
    it substracts/adds a tolerance from/to the min and max values. The tolerance
    is a fraction of the difference between times / interval of the MS (see
    get_utimes_tol()).

    Parameters
    ----------
    tb_tool : tables.table
        table (query) opened with an MSv4 query

    taql_where : str
        TaQL where that defines the partition of this MSv4

    Returns
    -------
    tuple
        min/max times (raw time values from the Msv2 table)
    """
    utimes, tol = get_utimes_tol(tb_tool, taql_where)
    time_min = utimes.min() - tol
    time_max = utimes.max() + tol
    return (time_min, time_max)


def create_data_variables(
    in_file, xds, tb_tool, time_baseline_shape, tidxs, bidxs, didxs, use_table_iter
):
    # Create Data Variables
    col_names = tb_tool.colnames()

    main_table_attrs = extract_table_attributes(in_file)
    main_column_descriptions = main_table_attrs["column_descriptions"]
    for col in col_names:
        if col in col_to_data_variable_names:
            if (col == "WEIGHT") and ("WEIGHT_SPECTRUM" in col_names):
                continue
            try:
                start = time.time()
                if col == "WEIGHT":
                    xds = get_weight(
                        xds,
                        col,
                        tb_tool,
                        time_baseline_shape,
                        tidxs,
                        bidxs,
                        use_table_iter,
                        main_column_descriptions,
                    )
                else:
                    xds[col_to_data_variable_names[col]] = xr.DataArray(
                        read_col_conversion(
                            tb_tool,
                            col,
                            time_baseline_shape,
                            tidxs,
                            bidxs,
                            use_table_iter,
                        ),
                        dims=col_dims[col],
                    )

                xds[col_to_data_variable_names[col]].attrs.update(
                    create_attribute_metadata(col, main_column_descriptions)
                )

                logger.debug(
                    "Time to read column " + str(col) + " : " + str(time.time() - start)
                )
            except Exception as exc:
                logger.debug(f"Could not load column {col}, exception: {exc}")

                if ("WEIGHT_SPECTRUM" == col) and (
                    "WEIGHT" in col_names
                ):  # Bogus WEIGHT_SPECTRUM column, need to use WEIGHT.
                    xds = get_weight(
                        xds,
                        "WEIGHT",
                        tb_tool,
                        time_baseline_shape,
                        tidxs,
                        bidxs,
                        use_table_iter,
                        main_column_descriptions,
                    )


def add_missing_data_var_attrs(xds):
    """
    Adds in the xds attributes expected metadata that cannot be found in the input MSv2.
    For now:
    - missing single-dish/SPECTRUM metadata
    - missing interferometry/VISIBILITY_MODEL metadata
    """
    data_var_names = ["SPECTRUM", "SPECTRUM_CORRECTED"]
    for var_name in data_var_names:
        if var_name in xds.data_vars:
            xds.data_vars[var_name].attrs["units"] = [""]

    vis_var_names = ["VISIBILITY_MODEL"]
    for var_name in vis_var_names:
        if var_name in xds.data_vars and "units" not in xds.data_vars[var_name].attrs:
            # Assume MODEL uses the same units
            if "VISIBILITY" in xds.data_vars:
                xds.data_vars[var_name].attrs["units"] = xds.data_vars[
                    "VISIBILITY"
                ].attrs["units"]
            else:
                xds.data_vars[var_name].attrs["units"] = [""]

    return xds


def get_weight(
    xds,
    col,
    tb_tool,
    time_baseline_shape,
    tidxs,
    bidxs,
    use_table_iter,
    main_column_descriptions,
):
    xds[col_to_data_variable_names[col]] = xr.DataArray(
        np.tile(
            read_col_conversion(
                tb_tool,
                col,
                time_baseline_shape,
                tidxs,
                bidxs,
                use_table_iter,
            )[:, :, None, :],
            (1, 1, xds.sizes["frequency"], 1),
        ),
        dims=col_dims[col],
    )

    xds[col_to_data_variable_names[col]].attrs.update(
        create_attribute_metadata(col, main_column_descriptions)
    )
    return xds


def create_taql_query_where(partition_info: dict):
    main_par_table_cols = [
        "DATA_DESC_ID",
        "OBSERVATION_ID",
        "STATE_ID",
        "FIELD_ID",
        "SCAN_NUMBER",
        "STATE_ID",
        "ANTENNA1",
    ]

    taql_where = "WHERE "
    for col_name in main_par_table_cols:
        if col_name in partition_info:
            taql_where = (
                taql_where
                + f"({col_name} IN [{','.join(map(str, partition_info[col_name]))}]) AND"
            )
            if col_name == "ANTENNA1":
                taql_where = (
                    taql_where
                    + f"(ANTENNA2 IN [{','.join(map(str, partition_info[col_name]))}]) AND"
                )
    taql_where = taql_where[:-3]

    return taql_where


def fix_uvw_frame(
    xds: xr.Dataset, field_and_source_xds: xr.Dataset, is_single_dish: bool
) -> xr.Dataset:
    """
    Fix UVW frame

    From CASA fixvis docs: clean and the im tool ignore the reference frame claimed by the UVW column (it is often
    mislabelled as ITRF when it is really FK5 (J2000)) and instead assume the (u, v, w)s are in the same frame as the phase
    tracking center. calcuvw does not yet force the UVW column and field centers to use the same reference frame!
    Blank = use the phase tracking frame of vis.
    """
    if xds.UVW.attrs["frame"] == "ITRF":
        if is_single_dish:
            center_var = "FIELD_REFERENCE_CENTER"
        else:
            center_var = "FIELD_PHASE_CENTER"

        xds.UVW.attrs["frame"] = field_and_source_xds[center_var].attrs["frame"]

    return xds


def estimate_memory_for_partition(in_file: str, partition: dict) -> float:
    """
    Aim: given a partition description, estimates a safe maximum memory value, but avoiding overestimation
    (at least not adding not well understood factors).
    """

    def calculate_term_all_data(
        tb_tool: tables.table, ntimes: float, nbaselines: float
    ) -> tuple[list[float], bool]:
        """
        Size that DATA vars from MS will have in the MSv4, whether this MS has FLOAT_DATA
        """
        sizes_all_data_vars = []
        col_names = tb_tool.colnames()
        for data_col in ["DATA", "CORRECTED_DATA", "MODEL_DATA", "FLOAT_DATA"]:
            if data_col in col_names:
                col_descr = tb_tool.getcoldesc(data_col)
                if "shape" in col_descr and isinstance(col_descr["shape"], np.ndarray):
                    # example: "shape": array([15,  4]) => gives pols x channels
                    cells_in_row = col_descr["shape"].prod()
                    npols = col_descr["shape"][-1]
                else:
                    first_row = np.array(tb_tool.col(data_col)[0])
                    cells_in_row = np.prod(first_row.shape)
                    npols = first_row.shape[-1]

                if col_descr["valueType"] == "complex":
                    # Assume. Otherwise, read first column and get the itemsize:
                    # col_dtype = np.array(mtable.col(data_col)[0]).dtype
                    # cell_size = col_dtype.itemsize
                    cell_size = 4
                    if data_col != "FLOAT_DATA":
                        cell_size *= 2
                elif col_descr["valueType"] == "float":
                    cell_size = 4

                # cells_in_row should account for the polarization and frequency dims
                size_data_var = ntimes * nbaselines * cells_in_row * cell_size

                sizes_all_data_vars.append(size_data_var)

        is_float_data = "FLOAT_DATA" in col_names

        return sizes_all_data_vars, is_float_data

    def calculate_term_weight_flag(size_largest_data, is_float_data) -> float:
        """
        Size that WEIGHT and FLAG will have in the MSv4, derived from the size of the
        MSv2 DATA col=> MSv4 VIS/SPECTRUM data var.
        """
        # Factors of the relative "cell_size" wrt the DATA var
        # WEIGHT_SPECTRUM size: DATA (IF), DATA/2 (SD)
        factor_weight = 1.0 if is_float_data else 0.5
        factor_flag = 1.0 / 4.0 if is_float_data else 1.0 / 8.0

        return size_largest_data * (factor_weight + factor_flag)

    def calculate_term_other_data_vars(
        ntimes: int, nbaselines: int, is_float_data: bool
    ) -> float:
        """
        Size all data vars other than the DATA (visibility/spectrum) vars will have in the MSv4

        For the rest of columns, including indices/iteration columns and other
        scalar columns could say approx ->5% of the (large) data cols

        """
        # Small ones, but as they are loaded into data arrays, why not including,
        # For example: UVW (3xscalar), EXPOSURE, TIME_CENTROID
        # assuming float64 in output MSv4
        item_size = 8
        return ntimes * nbaselines * (3 + 1 + 1) * item_size

    def calculate_term_calc_indx_for_row_split(msv2_nrows: int) -> float:
        """
        Account for the indices produced in calc_indx_for_row_split():
        the dominating ones are: tidxs, bidxs, didxs.

        In terms of amount of memory represented by this term relative to the
        total, it becomes relevant proportionally to the ratio between
           nrows / (chans x pols)
        - for example LOFAR long scans/partitions with few channels,
        but its value is independent from # chans, pols.
        """
        item_size = 8
        # 3 are: tidxs, bidxs, didxs
        return msv2_nrows * 3 * item_size

    def calculate_term_other_msv2_indices(msv2_nrows: int) -> float:
        """
        Account for the allocations to load ID, etc. columns from input MSv2.
        The converter needs to load: OBSERVATION_ID, INTERVAL, SCAN_NUMBER.
        These are loaded one after another (allocations do not stack up).
        Also, in most memory profiles these allocations are released once we
        get to create_data_variables(). As such, adding this term will most
        likely lead to overestimation (but adding it for safety).

        Simlarly as with calculate_term_calc_indx_for_row_split() this term
        becomes relevant when the ratio 'nrows / (chans x pols)' is high.
        """
        # assuming float64/int64 in input MSv2, which seems to be the case,
        # except for OBSERVATION_ID (int32)
        item_size = 8
        return msv2_nrows * item_size

    def calculate_term_attrs(size_estimate_main_xds: float) -> float:
        """Rough guess which seems to be more than enough"""
        # could also account for info_dicts (which seem to require typically ~1 MB)
        return 10 * 1024 * 1024

    def calculate_term_sub_xds(size_estimate_main_xds: float) -> float:
        """
        This is still very rough. Just seemingly working for now. Not taking into account the dims
        of the sub-xdss, interpolation options used, etc.
        """
        # Most cases so far 1% seems enough
        return 0.015 * size_estimate_main_xds

    def calculate_term_to_zarr(size_estimate_main_xds: float) -> float:
        """
        The to_zarr call on the main_xds seems to allocate 10s or 100s of MBs, presumably for buffers.
        That adds on top of the expected main_xds size.
        This is currently a very rough extrapolation and is being (mis)used to give a safe up to 5-6%
        overestimation. Perhaps we should drop this term once other sub-xdss are accounted for (and
        this term could be replaced by a similar, smaller but still safe over-estimation percentage).
        """
        return 0.05 * size_estimate_main_xds

    taql_partition = create_taql_query_where(partition)
    taql_main = f"select * from $mtable {taql_partition}"
    with open_table_ro(in_file) as mtable:
        col_names = mtable.colnames()
        with open_query(mtable, taql_main) as tb_tool:
            # Do not feel tempted to rely on nrows. nrows tends to underestimate memory when baselines are missing.
            # For some EVN datasets that can easily underestimate by a 50%
            utimes, _tol = get_utimes_tol(mtable, taql_partition)
            ntimes = len(utimes)
            nbaselines = len(get_baselines(tb_tool))

            # Still, use nrwos for estimations related to sizes of input (MSv2)
            # columns, not sizes of output (MSv4) data vars
            msv2_nrows = tb_tool.nrows()

            sizes_all_data, is_float_data = calculate_term_all_data(
                tb_tool, ntimes, nbaselines
            )

    size_largest_data = np.max(sizes_all_data)
    sum_sizes_data = np.sum(sizes_all_data)
    estimate_main_xds = (
        sum_sizes_data
        + calculate_term_weight_flag(size_largest_data, is_float_data)
        + calculate_term_other_data_vars(ntimes, nbaselines, is_float_data)
    )
    estimate = (
        estimate_main_xds
        + calculate_term_calc_indx_for_row_split(msv2_nrows)
        + calculate_term_other_msv2_indices(msv2_nrows)
        + calculate_term_sub_xds(estimate_main_xds)
        + calculate_term_to_zarr(estimate_main_xds)
    )
    estimate /= GiBYTES_TO_BYTES

    return estimate


def estimate_memory_and_cores_for_partitions(
    in_file: str, partitions: list
) -> tuple[float, int, int]:
    """
    Estimates approximate memory required to convert an MSv2 to MSv4, given
    a predefined set of partitions.
    """
    max_cores = len(partitions)

    size_estimates = [
        estimate_memory_for_partition(in_file, part_description)
        for part_description in partitions
    ]
    max_estimate = np.max(size_estimates) if size_estimates else 0.0

    recommended_cores = np.ceil(max_cores / 4).astype("int")

    return float(max_estimate), int(max_cores), int(recommended_cores)


def convert_and_write_partition(
    in_file: str,
    out_file: str,
    ms_v4_id: Union[int, str],
    partition_info: Dict,
    use_table_iter: bool,
    partition_scheme: str = "ddi_intent_field",
    main_chunksize: Union[Dict, float, None] = None,
    with_pointing: bool = True,
    pointing_chunksize: Union[Dict, float, None] = None,
    pointing_interpolate: bool = False,
    ephemeris_interpolate: bool = False,
    phase_cal_interpolate: bool = False,
    sys_cal_interpolate: bool = False,
    compressor: numcodecs.abc.Codec = numcodecs.Zstd(level=2),
    storage_backend="zarr",
    overwrite: bool = False,
):
    """_summary_

    Parameters
    ----------
    in_file : str
        _description_
    out_file : str
        _description_
    intents : str
        _description_
    ddi : int, optional
        _description_, by default 0
    state_ids : _type_, optional
        _description_, by default None
    field_id : int, optional
        _description_, by default None
    main_chunksize : Union[Dict, float, None], optional
        _description_, by default None
    with_pointing: bool, optional
        _description_, by default True
    pointing_chunksize : Union[Dict, float, None], optional
        _description_, by default None
    pointing_interpolate : bool, optional
        _description_, by default None
    ephemeris_interpolate : bool, optional
        _description_, by default None
    phase_cal_interpolate : bool, optional
        _description_, by default None
    sys_cal_interpolate : bool, optional
        _description_, by default None
    compressor : numcodecs.abc.Codec, optional
        _description_, by default numcodecs.Zstd(level=2)
    storage_backend : str, optional
        _description_, by default "zarr"
    overwrite : bool, optional
        _description_, by default False

    Returns
    -------
    _type_
        _description_
    """

    taql_where = create_taql_query_where(partition_info)
    ddi = partition_info["DATA_DESC_ID"][0]
    intents = str(partition_info["OBS_MODE"][0])

    start = time.time()
    with open_table_ro(in_file) as mtable:
        taql_main = f"select * from $mtable {taql_where}"
        with open_query(mtable, taql_main) as tb_tool:

            if tb_tool.nrows() == 0:
                tb_tool.close()
                mtable.close()
                return xr.Dataset(), {}, {}

            logger.debug("Starting a real convert_and_write_partition")
            (
                tidxs,
                bidxs,
                didxs,
                baseline_ant1_id,
                baseline_ant2_id,
                utime,
            ) = calc_indx_for_row_split(tb_tool, taql_where)
            time_baseline_shape = (len(utime), len(baseline_ant1_id))
            logger.debug("Calc indx for row split " + str(time.time() - start))

            observation_id = check_if_consistent(
                tb_tool.getcol("OBSERVATION_ID"), "OBSERVATION_ID"
            )

            def get_observation_info(in_file, observation_id, intents):
                generic_observation_xds = load_generic_table(
                    in_file,
                    "OBSERVATION",
                    taql_where=f" where (ROWID() IN [{str(observation_id)}])",
                )

                if intents == "None":
                    intents = "obs_" + str(observation_id)

                return generic_observation_xds["TELESCOPE_NAME"].values[0], intents

            telescope_name, intents = get_observation_info(
                in_file, observation_id, intents
            )

            start = time.time()
            xds = xr.Dataset(
                attrs={
                    "creation_date": datetime.datetime.now(
                        datetime.timezone.utc
                    ).isoformat(),
                    "xradio_version": importlib.metadata.version("xradio"),
                    "schema_version": "4.0.-9994",
                    "type": "visibility",
                }
            )

            # interval = check_if_consistent(tb_tool.getcol("INTERVAL"), "INTERVAL")
            interval = tb_tool.getcol("INTERVAL")

            interval_unique = unique_1d(interval)
            if len(interval_unique) > 1:
                logger.debug(
                    "Integration time (interval) not consitent in partition, using median."
                )
                interval = np.median(interval)
            else:
                interval = interval_unique[0]

            scan_id = np.full(time_baseline_shape, -42, dtype=int)
            scan_id[tidxs, bidxs] = tb_tool.getcol("SCAN_NUMBER")
            scan_id = np.max(scan_id, axis=1)

            xds = create_coordinates(
                xds,
                in_file,
                ddi,
                utime,
                interval,
                baseline_ant1_id,
                baseline_ant2_id,
                scan_id,
            )
            logger.debug("Time create coordinates " + str(time.time() - start))

            start = time.time()
            create_data_variables(
                in_file,
                xds,
                tb_tool,
                time_baseline_shape,
                tidxs,
                bidxs,
                didxs,
                use_table_iter,
            )

            # Add data_groups
            xds, is_single_dish = add_data_groups(xds)
            xds = add_missing_data_var_attrs(xds)

            if (
                "WEIGHT" not in xds.data_vars
            ):  # Some single dish datasets don't have WEIGHT.
                if is_single_dish:
                    xds["WEIGHT"] = xr.DataArray(
                        np.ones(xds.SPECTRUM.shape, dtype=np.float64),
                        dims=xds.SPECTRUM.dims,
                    )
                else:
                    xds["WEIGHT"] = xr.DataArray(
                        np.ones(xds.VISIBILITY.shape, dtype=np.float64),
                        dims=xds.VISIBILITY.dims,
                    )

            logger.debug("Time create data variables " + str(time.time() - start))

            # To constrain the time range to load (in pointing, ephemerides, phase_cal data_vars)
            time_min_max = find_min_max_times(tb_tool, taql_where)

            # Create ant_xds
            start = time.time()
            feed_id = unique_1d(
                np.concatenate(
                    [
                        unique_1d(tb_tool.getcol("FEED1")),
                        unique_1d(tb_tool.getcol("FEED2")),
                    ]
                )
            )
            antenna_id = unique_1d(
                np.concatenate(
                    [xds["baseline_antenna1_id"].data, xds["baseline_antenna2_id"].data]
                )
            )
            if phase_cal_interpolate:
                phase_cal_interp_time = xds.time.values
            else:
                phase_cal_interp_time = None

            ant_xds = create_antenna_xds(
                in_file,
                xds.frequency.attrs["spectral_window_id"],
                antenna_id,
                feed_id,
                telescope_name,
                xds.polarization,
            )
            logger.debug("Time antenna xds  " + str(time.time() - start))

            start = time.time()
            gain_curve_xds = create_gain_curve_xds(
                in_file, xds.frequency.attrs["spectral_window_id"], ant_xds
            )
            logger.debug("Time gain_curve xds  " + str(time.time() - start))

            start = time.time()
            phase_calibration_xds = create_phase_calibration_xds(
                in_file,
                xds.frequency.attrs["spectral_window_id"],
                ant_xds,
                time_min_max,
                phase_cal_interp_time,
            )
            logger.debug("Time phase_calibration xds  " + str(time.time() - start))

            # Change antenna_ids to antenna_names
            with_antenna_partitioning = "ANTENNA1" in partition_info
            xds = antenna_ids_to_names(
                xds, ant_xds, is_single_dish, with_antenna_partitioning
            )
            # but before, keep the name-id arrays, we need them for the pointing and weather xds
            ant_xds_name_ids = ant_xds["antenna_name"].set_xindex("antenna_id")
            ant_xds_station_name_ids = ant_xds["station"].set_xindex("antenna_id")
            # No longer needed after converting to name.
            ant_xds = ant_xds.drop_vars("antenna_id")

            # Create system_calibration_xds
            start = time.time()
            if sys_cal_interpolate:
                sys_cal_interp_time = xds.time.values
            else:
                sys_cal_interp_time = None
            system_calibration_xds = create_system_calibration_xds(
                in_file,
                xds.frequency,
                ant_xds_name_ids,
                sys_cal_interp_time,
            )
            logger.debug("Time system_calibation " + str(time.time() - start))

            # Create weather_xds
            start = time.time()
            weather_xds = create_weather_xds(in_file, ant_xds_station_name_ids)
            logger.debug("Time weather " + str(time.time() - start))

            # Create pointing_xds
            pointing_xds = xr.Dataset()
            if with_pointing:
                start = time.time()
                if pointing_interpolate:
                    pointing_interp_time = xds.time
                else:
                    pointing_interp_time = None
                pointing_xds = create_pointing_xds(
                    in_file, ant_xds_name_ids, time_min_max, pointing_interp_time
                )
                pointing_chunksize = parse_chunksize(
                    pointing_chunksize, "pointing", pointing_xds
                )
                add_encoding(
                    pointing_xds, compressor=compressor, chunks=pointing_chunksize
                )
                logger.debug(
                    "Time pointing (with add compressor and chunking) "
                    + str(time.time() - start)
                )

            start = time.time()

            # Time and frequency should always be increasing
            if len(xds.frequency) > 1 and xds.frequency[1] - xds.frequency[0] < 0:
                xds = xds.sel(frequency=slice(None, None, -1))

            if len(xds.time) > 1 and xds.time[1] - xds.time[0] < 0:
                xds = xds.sel(time=slice(None, None, -1))

            # Create field_and_source_xds (combines field, source and ephemeris data into one super dataset)
            start = time.time()
            if ephemeris_interpolate:
                ephemeris_interp_time = xds.time.values
            else:
                ephemeris_interp_time = None

            if "FIELD_ID" not in partition_scheme:
                field_id = np.full(time_baseline_shape, -42, dtype=int)
                field_id[tidxs, bidxs] = tb_tool.getcol("FIELD_ID")
                field_id = np.max(field_id, axis=1)
                field_times = utime
            else:
                field_id = check_if_consistent(tb_tool.getcol("FIELD_ID"), "FIELD_ID")
                field_times = None

            # col_unique = unique_1d(col)
            # assert len(col_unique) == 1, col_name + " is not consistent."
            # return col_unique[0]

            field_and_source_xds, source_id, _num_lines = create_field_and_source_xds(
                in_file,
                field_id,
                xds.frequency.attrs["spectral_window_id"],
                field_times,
                is_single_dish,
                time_min_max,
                ephemeris_interp_time,
            )
            logger.debug("Time field_and_source_xds " + str(time.time() - start))

            xds = fix_uvw_frame(xds, field_and_source_xds, is_single_dish)

            partition_info_misc_fields = {
                "scan_id": scan_id,
                "intents": intents,
                "taql_where": taql_where,
            }
            info_dicts = create_info_dicts(
                in_file, xds, field_and_source_xds, partition_info_misc_fields, tb_tool
            )
            xds.attrs.update(info_dicts)

            # xds ready, prepare to write
            start = time.time()
            main_chunksize = parse_chunksize(main_chunksize, "main", xds)
            add_encoding(xds, compressor=compressor, chunks=main_chunksize)
            logger.debug("Time add compressor and chunk " + str(time.time() - start))

            file_name = os.path.join(
                out_file,
                pathlib.Path(in_file).name.replace(".ms", "") + "_" + str(ms_v4_id),
            )

            if overwrite:
                mode = "w"
            else:
                mode = "w-"

            if is_single_dish:
                xds.attrs["type"] = "spectrum"
                xds = xds.drop_vars(["UVW"])
                del xds["uvw_label"]
            else:
                if any("WVR" in s for s in intents):
                    xds.attrs["type"] = "wvr"
                else:
                    xds.attrs["type"] = "visibility"

            import sys

            start = time.time()
            if storage_backend == "zarr":
                xds.to_zarr(store=os.path.join(file_name, "correlated_xds"), mode=mode)
                ant_xds.to_zarr(store=os.path.join(file_name, "antenna_xds"), mode=mode)
                for group_name in xds.attrs["data_groups"]:
                    field_and_source_xds.to_zarr(
                        store=os.path.join(
                            file_name, f"field_and_source_xds_{group_name}"
                        ),
                        mode=mode,
                    )

                if with_pointing and len(pointing_xds.data_vars) > 0:
                    pointing_xds.to_zarr(
                        store=os.path.join(file_name, "pointing_xds"), mode=mode
                    )

                if system_calibration_xds:
                    system_calibration_xds.to_zarr(
                        store=os.path.join(file_name, "system_calibration_xds"),
                        mode=mode,
                    )

                if gain_curve_xds:
                    gain_curve_xds.to_zarr(
                        store=os.path.join(file_name, "gain_curve_xds"), mode=mode
                    )

                if phase_calibration_xds:
                    phase_calibration_xds.to_zarr(
                        store=os.path.join(file_name, "phase_calibration_xds"),
                        mode=mode,
                    )

                if weather_xds:
                    weather_xds.to_zarr(
                        store=os.path.join(file_name, "weather_xds"), mode=mode
                    )

            elif storage_backend == "netcdf":
                # xds.to_netcdf(path=file_name+"/MAIN", mode=mode) #Does not work
                raise
            logger.debug("Write data  " + str(time.time() - start))

    # logger.info("Saved ms_v4 " + file_name + " in " + str(time.time() - start_with) + "s")


def antenna_ids_to_names(
    xds: xr.Dataset,
    ant_xds: xr.Dataset,
    is_single_dish: bool,
    with_antenna_partitioning,
) -> xr.Dataset:
    """
    Turns the antenna_ids that we get from MSv2 into MSv4 antenna_name

    Parameters
    ----------
    xds: xr.Dataset
        A main xds (MSv4)
    ant_xds: xr.Dataset
        The antenna_xds for this MSv4
    is_single_dish: bool
        Whether a single-dish ("spectrum" data) dataset
    with_antenna_partitioning: bool
        Whether the MSv4 partitions include the antenna axis => only
        one antenna (and implicitly one 'baseline' - auto-correlation)

    Returns
    ----------
    xr.Dataset
        The main xds with antenna_id replaced with antenna_name
    """
    ant_xds = ant_xds.set_xindex(
        "antenna_id"
    )  # Allows for non-dimension coordinate selection.

    if not is_single_dish:  # Interferometer
        xds["baseline_antenna1_id"].data = ant_xds["antenna_name"].sel(
            antenna_id=xds["baseline_antenna1_id"].data
        )
        xds["baseline_antenna2_id"].data = ant_xds["antenna_name"].sel(
            antenna_id=xds["baseline_antenna2_id"].data
        )
        xds = xds.rename(
            {
                "baseline_antenna1_id": "baseline_antenna1_name",
                "baseline_antenna2_id": "baseline_antenna2_name",
            }
        )
    else:
        if not with_antenna_partitioning:
            # baseline_antenna1_id will be removed soon below, but it is useful here to know the actual antenna_ids,
            # as opposed to the baseline_ids which can mismatch when data is missing for some antennas
            xds["baseline_id"] = ant_xds["antenna_name"].sel(
                antenna_id=xds["baseline_antenna1_id"]
            )
        else:
            xds["baseline_id"] = ant_xds["antenna_name"]

        unwanted_coords_from_ant_xds = [
            "antenna_id",
            "antenna_name",
            "mount",
            "station",
        ]
        for unwanted_coord in unwanted_coords_from_ant_xds:
            xds = xds.drop_vars(unwanted_coord)

        # Rename a dim coord started generating warnings (index not re-created). Swap dims, create coord
        # https://github.com/pydata/xarray/pull/6999
        xds = xds.swap_dims({"baseline_id": "antenna_name"})
        xds = xds.assign_coords({"antenna_name": xds["baseline_id"].data})
        xds = xds.drop_vars("baseline_id")

        # drop more vars that seem unwanted in main_sd_xds, but there shouuld be a better way
        # of not creating them in the first place
        unwanted_coords_sd = ["baseline_antenna1_id", "baseline_antenna2_id"]
        for unwanted_coord in unwanted_coords_sd:
            xds = xds.drop_vars(unwanted_coord)

    return xds


def add_data_groups(xds):
    xds.attrs["data_groups"] = {}
    if "VISIBILITY" in xds:
        xds.attrs["data_groups"]["base"] = {
            "correlated_data": "VISIBILITY",
            "flag": "FLAG",
            "weight": "WEIGHT",
            "uvw": "UVW",
        }

    if "VISIBILITY_CORRECTED" in xds:
        xds.attrs["data_groups"]["corrected"] = {
            "correlated_data": "VISIBILITY_CORRECTED",
            "flag": "FLAG",
            "weight": "WEIGHT",
            "uvw": "UVW",
        }

    if "VISIBILITY_MODEL" in xds:
        xds.attrs["data_groups"]["model"] = {
            "correlated_data": "VISIBILITY_MODEL",
            "flag": "FLAG",
            "weight": "WEIGHT",
            "uvw": "UVW",
        }

    is_single_dish = False
    if "SPECTRUM" in xds:
        xds.attrs["data_groups"]["base"] = {
            "correlated_data": "SPECTRUM",
            "flag": "FLAG",
            "weight": "WEIGHT",
            "uvw": "UVW",
        }
        is_single_dish = True

    if "SPECTRUM_MODEL" in xds:
        xds.attrs["data_groups"]["model"] = {
            "correlated_data": "SPECTRUM_MODEL",
            "flag": "FLAG",
            "weight": "WEIGHT",
            "uvw": "UVW",
        }
        is_single_dish = True

    if "SPECTRUM_CORRECTED" in xds:
        xds.attrs["data_groups"]["corrected"] = {
            "correlated_data": "SPECTRUM_CORRECTED",
            "flag": "FLAG",
            "weight": "WEIGHT",
            "uvw": "UVW",
        }
        is_single_dish = True

    return xds, is_single_dish



import toolviper.utils.logger as logger
import numcodecs
from typing import Dict, Union

import dask

from xradio.measurement_set._utils._msv2.partition_queries import create_partitions
from xradio.measurement_set._utils._msv2.conversion import (
    convert_and_write_partition,
    estimate_memory_and_cores_for_partitions,
)


def estimate_conversion_memory_and_cores(
    in_file: str,
    partition_scheme: list = ["FIELD_ID"],
) -> tuple[float, int, int]:
    """
    Given an MSv2 and a partition_scheme to use when converting it to MSv4,
    estimates:
    - memory (in the sense of the amount expected to be enough to convert)
    - cores (in the sense of the recommended/optimal number of cores to use to convert)

    Note: this function does not currently try to estimate the memory required for
    sub-xdss such as pointing_xds and system_calibration_xds, instead it uses a small
    percentage of the main_xds to account for them. This can lead to underestimation
    especially for MSv2s with small partitions but large pointing or syscal tables.
    This should not typically be a concern for sufficiently large partitions
    (a few or 10s, 100s of GiBs).

    Parameters
    ----------
    in_file: str
        Input MS name.
    partition_scheme: list
        Partition scheme as used in the function convert_msv2_to_processing_set()

    Returns
    ----------
    tuple
        estimated maximum memory required for one partition,
        maximum number of cores it makes sense to use (number of partitions),
        suggested number of cores to use (maximum/4 as a rule of thumb)
    """

    partitions = create_partitions(in_file, partition_scheme=partition_scheme)

    return estimate_memory_and_cores_for_partitions(in_file, partitions)


def convert_msv2_to_processing_set(
    in_file: str,
    out_file: str,
    partition_scheme: list = ["FIELD_ID"],
    main_chunksize: Union[Dict, float, None] = None,
    with_pointing: bool = True,
    pointing_chunksize: Union[Dict, float, None] = None,
    pointing_interpolate: bool = False,
    ephemeris_interpolate: bool = False,
    phase_cal_interpolate: bool = False,
    sys_cal_interpolate: bool = False,
    use_table_iter: bool = False,
    compressor: numcodecs.abc.Codec = numcodecs.Zstd(level=2),
    storage_backend: str = "zarr",
    parallel: bool = False,
    overwrite: bool = False,
):
    """Convert a Measurement Set v2 into a Processing Set of Measurement Set v4.

    Parameters
    ----------
    in_file : str
        Input MS name.
    out_file : str
        Output PS name.
    partition_scheme : list, optional
        A MS v4 can only contain a single data description (spectral window and polarization setup), and observation mode. Consequently, the MS v2 is partitioned when converting to MS v4.
        In addition to data description and polarization setup a finer partitioning is possible by specifying a list of partitioning keys. Any combination of the following keys are possible:
        "FIELD_ID", "SCAN_NUMBER", "STATE_ID", "SOURCE_ID", "SUB_SCAN_NUMBER". For mosaics where the phase center is rapidly changing (such as VLA on the fly mosaics)
        partition_scheme should be set to an empty list []. By default, ["FIELD_ID"].
    main_chunksize : Union[Dict, float, None], optional
        Defines the chunk size of the main dataset. If given as a dictionary, defines the sizes of several dimensions, and acceptable keys are "time", "baseline_id", "antenna_id", "frequency", "polarization". If given as a float, gives the size of a chunk in GiB. By default, None.
    with_pointing : bool, optional
        Whether to convert the POINTING subtable into pointing sub-datasets
    pointing_chunksize : Union[Dict, float, None], optional
        Defines the chunk size of the pointing dataset. If given as a dictionary, defines the sizes of several dimensions, acceptable keys are "time" and "antenna_id". If given as a float, defines the size of a chunk in GiB. By default, None.
    pointing_interpolate : bool, optional
        Whether to interpolate the time axis of the pointing sub-dataset to the time axis of the main dataset
    ephemeris_interpolate : bool, optional
        Whether to interpolate the time axis of the ephemeris data variables (of the field_and_source sub-dataset) to the time axis of the main dataset
    phase_cal_interpolate : bool, optional
        Whether to interpolate the time axis of the phase calibration data variables to the time axis of the main dataset
    sys_cal_interpolate : bool, optional
        Whether to interpolate the time axis of the system calibration data variables (sys_cal_xds) to the time axis of the main dataset
    use_table_iter : bool, optional
        Whether to use the table iterator to read the main table of the MS v2. This should be set to True when reading datasets with large number of rows and few partitions, by default False.
    compressor : numcodecs.abc.Codec, optional
        The Blosc compressor to use when saving the converted data to disk using Zarr, by default numcodecs.Zstd(level=2).
    storage_backend : {"zarr", "netcdf"}, optional
        The on-disk format to use. "netcdf" is not yet implemented.
    parallel : bool, optional
        Makes use of Dask to execute conversion in parallel, by default False.
    overwrite : bool, optional
        Whether to overwrite an existing processing set, by default False.
    """

    partitions = create_partitions(in_file, partition_scheme=partition_scheme)
    logger.info("Number of partitions: " + str(len(partitions)))
    delayed_list = []

    for ms_v4_id, partition_info in enumerate(partitions):
        # print(ms_v4_id,len(partition_info['FIELD_ID']))

        logger.info(
            "OBSERVATION_ID "
            + str(partition_info["OBSERVATION_ID"])
            + ", DDI "
            + str(partition_info["DATA_DESC_ID"])
            + ", STATE "
            + str(partition_info["STATE_ID"])
            + ", FIELD "
            + str(partition_info["FIELD_ID"])
            + ", SCAN "
            + str(partition_info["SCAN_NUMBER"])
            + (
                ", ANTENNA " + str(partition_info["ANTENNA1"])
                if "ANTENNA1" in partition_info
                else ""
            )
        )

        # prepend '0' to ms_v4_id as needed
        ms_v4_id = f"{ms_v4_id:0>{len(str(len(partitions) - 1))}}"
        if parallel:
            delayed_list.append(
                dask.delayed(convert_and_write_partition)(
                    in_file,
                    out_file,
                    ms_v4_id,
                    partition_info=partition_info,
                    use_table_iter=use_table_iter,
                    partition_scheme=partition_scheme,
                    main_chunksize=main_chunksize,
                    with_pointing=with_pointing,
                    pointing_chunksize=pointing_chunksize,
                    pointing_interpolate=pointing_interpolate,
                    ephemeris_interpolate=ephemeris_interpolate,
                    phase_cal_interpolate=phase_cal_interpolate,
                    sys_cal_interpolate=sys_cal_interpolate,
                    compressor=compressor,
                    overwrite=overwrite,
                )
            )
        else:
            convert_and_write_partition(
                in_file,
                out_file,
                ms_v4_id,
                partition_info=partition_info,
                use_table_iter=use_table_iter,
                partition_scheme=partition_scheme,
                main_chunksize=main_chunksize,
                with_pointing=with_pointing,
                pointing_chunksize=pointing_chunksize,
                pointing_interpolate=pointing_interpolate,
                ephemeris_interpolate=ephemeris_interpolate,
                phase_cal_interpolate=phase_cal_interpolate,
                sys_cal_interpolate=sys_cal_interpolate,
                compressor=compressor,
                overwrite=overwrite,
            )

    if parallel:
        dask.compute(delayed_list)



import os
import numpy as np
import json
import zarr
import s3fs
from xradio._utils.zarr.common import _get_file_system_and_items

from numcodecs.compat import (
    ensure_text,
    ensure_ndarray_like,
    ensure_bytes,
    ensure_contiguous_ndarray_like,
)

full_dims_lm = ["time", "polarization", "frequency", "l", "m"]
full_dims_uv = ["time", "polarization", "frequency", "l", "m"]
norm_dims = ["polarization", "frequency"]

image_data_variables_and_dims_double_precision = {
    "aperture": {"dims": full_dims_uv, "dtype": "<c16", "name": "APERTURE"},
    "aperture_normalization": {
        "dims": norm_dims,
        "dtype": "<c16",
        "name": "APERTURE_NORMALIZATION",
    },
    "primary_beam": {"dims": full_dims_lm, "dtype": "<f8", "name": "PRIMARY_BEAM"},
    "uv_sampling": {"dims": full_dims_uv, "dtype": "<c16", "name": "UV_SAMPLING"},
    "uv_sampling_normalization": {
        "dims": norm_dims,
        "dtype": "<c16",
        "name": "UV_SAMPLING_NORMALIZATION",
    },
    "point_spread_function": {
        "dims": full_dims_lm,
        "dtype": "<f8",
        "name": "POINT_SPREAD_FUNCTION",
    },
    "visibility": {"dims": full_dims_uv, "dtype": "<c16", "name": "VISIBILITY"},
    "visibility_normalization": {
        "dims": norm_dims,
        "dtype": "<c16",
        "name": "VISIBILITY_NORMALIZATION",
    },
    "sky": {"dims": full_dims_lm, "dtype": "<f8", "name": "SKY"},
}

image_data_variables_and_dims_single_precision = {
    "aperture": {"dims": full_dims_uv, "dtype": "<c8", "name": "APERTURE"},
    "aperture_normalization": {
        "dims": norm_dims,
        "dtype": "<c16",
        "name": "APERTURE_NORMALIZATION",
    },
    "primary_beam": {"dims": full_dims_lm, "dtype": "<f4", "name": "PRIMARY_BEAM"},
    "uv_sampling": {"dims": full_dims_uv, "dtype": "<c8", "name": "UV_SAMPLING"},
    "uv_sampling_normalization": {
        "dims": norm_dims,
        "dtype": "<c16",
        "name": "UV_SAMPLING_NORMALIZATION",
    },
    "point_spread_function": {
        "dims": full_dims_lm,
        "dtype": "<f8",
        "name": "POINT_SPREAD_FUNCTION",
    },
    "visibility": {"dims": full_dims_uv, "dtype": "<c8", "name": "VISIBILITY"},
    "visibility_normalization": {
        "dims": norm_dims,
        "dtype": "<c16",
        "name": "VISIBILITY_NORMALIZATION",
    },
    "sky": {"dims": full_dims_lm, "dtype": "<f4", "name": "SKY"},
}


def pad_array_with_nans(input_array, output_shape, dtype):
    """
    Pad an integer array with NaN values to match the specified output shape.

    Parameters:
    - input_array: The input NumPy array to be padded.
    - output_shape: A tuple specifying the desired output shape (e.g., (rows, columns)).

    Returns:
    - A NumPy array with NaN padding to match the specified output shape.
    """
    # Get the input shape
    input_shape = input_array.shape

    # Calculate the padding dimensions
    padding_shape = tuple(max(0, o - i) for i, o in zip(input_shape, output_shape))

    # Create a new array filled with NaN values
    padded_array = np.empty(output_shape, dtype=dtype)
    padded_array[:] = np.nan

    # Copy the input array to the appropriate position within the padded array
    padded_array[: input_shape[0], : input_shape[1], : input_shape[2]] = input_array

    return padded_array


def write_binary_blob_to_disk(arr, file_path, compressor):
    """
    Compress a NumPy array using Blosc and save it to disk.

    Parameters:
    - arr: The NumPy array to compress and save.
    - file_path: The path to the output file where the compressed array will be saved.
    - compressor:

    Returns:
    - None
    """
    import toolviper.utils.logger as logger

    # Encode the NumPy array using the codec
    logger.debug("1. Before compressor " + file_path)
    compressed_arr = compressor.encode(np.ascontiguousarray(arr))

    logger.debug("2. Before makedir")
    # Ensure the directory exists before saving the file
    os.makedirs(os.path.dirname(file_path), exist_ok=True)

    arr_len = len(compressed_arr)
    logger.debug("3. Before write the len is: " + str(arr_len))
    # Save the compressed array to disk
    # with open(file_path, "wb") as file:
    #     file.write(compressed_arr)

    logger.debug("4. Using new writer: " + str(arr_len))
    write_to_lustre_chunked(file_path, compressed_arr)

    # /.lustre/aoc/sciops/pford/CHILES/cube_image/uid___A002_Xee7674_X2844_Cube_3.img.zarr/SKY/0.0.110.0.0
    # 348192501 bytes
    # 332.0622453689575 M

    # from io import BufferedWriter
    # # Calculate buffer size based on compressed_arr size (adjust multiplier)
    # buffer_size = min(len(compressed_arr), 1024 * 1024 * 4)  # Max 4 MB buffer
    # with BufferedWriter(open(file_path, "wb"), buffer_size) as f:
    #     f.write(compressed_arr)
    #     f.flush()  # Ensure data gets written to disk

    logger.debug("4. Write completed")

    # print(f"Compressed array saved to {file_path}")


def write_to_lustre_chunked(
    file_path,
    compressed_arr,
    chunk_size=1024 * 1024 * 128,
):  # 128 MiB chunks
    """
    Writes compressed data to a Lustre file path with chunking.

    Args:
        file_path: Path to the file for writing.
        compressed_arr: Compressed data array to write.
        chunk_size: Size of each data chunk in bytes (default: 128 MiB).
    """
    fs, items = _get_file_system_and_items(file_path.rsplit("/", 1)[0])

    if isinstance(fs, s3fs.core.S3FileSystem):
        with fs.open(file_path, "wb") as f:
            for i in range(0, len(compressed_arr), chunk_size):
                chunk = compressed_arr[i : i + chunk_size]
                f.write(chunk)
    else:
        with open(file_path, "wb") as f:
            for i in range(0, len(compressed_arr), chunk_size):
                chunk = compressed_arr[i : i + chunk_size]
                f.write(chunk)


def read_binary_blob_from_disk(file_path, compressor, dtype=np.float64):
    """
    Read a compressed binary blob from disk and decode it using Blosc.

    Parameters:
    - file_path: The path to the compressed binary blob file.
    - compressor: The Blosc compressor to use (e.g., 'zstd', 'lz4', 'blosclz', etc.).

    Returns:
    - The decoded NumPy array.
    """

    # Check if the file exists
    if not os.path.exists(file_path):
        print(f"Error: File '{file_path}' not found.")
        return None

    # Read the compressed binary blob from disk
    with open(file_path, "rb") as file:
        compressed_arr = file.read()

    # Decode the compressed data using the Blosc compressor
    decoded_bytes = compressor.decode(compressed_arr)

    decoded_arr = np.frombuffer(decoded_bytes, dtype)  # Adjust dtype as needed

    return decoded_arr


def read_json_file(file_path):
    """
    Read a JSON file and return its contents as a Python dictionary.

    Parameters:
    - file_path: The path to the JSON file to be read.

    Returns:
    - A Python dictionary containing the JSON data.
    """
    try:
        with open(file_path, "r") as file:
            data = json.load(file)
        return data
    except FileNotFoundError:
        print(f"Error: File '{file_path}' not found.")
        return None
    except json.JSONDecodeError as e:
        print(f"Error decoding JSON from '{file_path}': {e}")
        return None


class NumberEncoder(json.JSONEncoder):
    def default(self, o):
        # See json.JSONEncoder.default docstring for explanation
        # This is necessary to encode numpy dtype
        if isinstance(o, numbers.Integral):
            return int(o)
        if isinstance(o, numbers.Real):
            return float(o)
        return json.JSONEncoder.default(self, o)


def write_json_file(data, file_path):
    """
    Write a Python dictionary to a JSON file.

    Parameters:
    - data: The Python dictionary to be written to the JSON file.
    - file_path: The path to the JSON file to be created or overwritten.

    Returns:
    - None
    """

    with open(file_path, "w") as file:
        json.dump(
            data,
            file,
            indent=4,
            sort_keys=True,
            ensure_ascii=True,
            separators=(",", ": "),
            cls=NumberEncoder,
        )


def create_data_variable_meta_data(
    zarr_group_name, data_variables_and_dims, xds_dims, parallel_coords, compressor
):
    zarr_meta = data_variables_and_dims

    fs, items = _get_file_system_and_items(zarr_group_name)

    for data_variable_key, dims_dtype_name in data_variables_and_dims.items():
        # print(data_variable_key, dims_dtype_name)

        dims = dims_dtype_name["dims"]
        dtype = dims_dtype_name["dtype"]
        data_variable_name = dims_dtype_name["name"]

        data_variable_path = os.path.join(zarr_group_name, data_variable_name)
        if isinstance(fs, s3fs.core.S3FileSystem):
            # N.b.,stateful "folder creation" is not a well defined concept for S3 objects and URIs
            # see https://github.com/fsspec/s3fs/issues/401
            # nor is a path specifier (cf. "URI")
            fs.mkdir(data_variable_path)
        else:
            # default to assuming we can use the os module and mkdir system call
            os.system("mkdir " + data_variable_path)
        # Create .zattrs
        zattrs = {
            "_ARRAY_DIMENSIONS": dims,
            # "coordinates": "time declination right_ascension"
        }

        shape = []
        chunks = []
        for d in dims:
            shape.append(xds_dims[d])
            if d in parallel_coords:
                chunks.append(len(parallel_coords[d]["data_chunks"][0]))
            else:
                chunks.append(xds_dims[d])

        # print(chunks,shape)
        # assuming data_variable_path has been set compatibly
        zattrs_file = os.path.join(data_variable_path, ".zattrs")

        if isinstance(fs, s3fs.core.S3FileSystem):
            with fs.open(zattrs_file, "w") as file:
                json.dump(
                    zattrs,
                    file,
                    indent=4,
                    sort_keys=True,
                    ensure_ascii=True,
                    separators=(",", ": "),
                    cls=NumberEncoder,
                )
        else:
            # default to assuming we can use primitives
            write_json_file(zattrs, zattrs_file)

        # Create .zarray
        from zarr import n5

        compressor_config = n5.compressor_config_to_zarr(
            n5.compressor_config_to_n5(compressor.get_config())
        )

        if "f" in dtype:
            zarray = {
                "chunks": chunks,
                "compressor": compressor_config,
                "dtype": dtype,
                "fill_value": "NaN",
                "filters": None,
                "order": "C",
                "shape": shape,
                "zarr_format": 2,
            }

        else:
            zarray = {
                "chunks": chunks,
                "compressor": compressor_config,
                "dtype": dtype,
                "fill_value": None,
                "filters": None,
                "order": "C",
                "shape": shape,
                "zarr_format": 2,
            }

        zarr_meta[data_variable_key]["chunks"] = chunks
        zarr_meta[data_variable_key]["shape"] = shape

        # again, assuming data_variable_path has been set compatibly
        zarray_file = os.path.join(data_variable_path, ".zarray")

        if isinstance(fs, s3fs.core.S3FileSystem):
            with fs.open(zarray_file, "w") as file:
                json.dump(
                    zarray,
                    file,
                    indent=4,
                    sort_keys=True,
                    ensure_ascii=True,
                    separators=(",", ": "),
                    cls=NumberEncoder,
                )
        else:
            # default to assuming we can use primitives
            write_json_file(zarray, zarray_file)

    return zarr_meta


def write_chunk(img_xds, meta, parallel_dims_chunk_id, compressor, image_file):
    dims = meta["dims"]
    dtype = meta["dtype"]
    data_variable_name = meta["name"]
    chunks = meta["chunks"]
    shape = meta["shape"]
    chunk_name = ""
    if data_variable_name in img_xds:
        for d in img_xds[data_variable_name].dims:
            if d in parallel_dims_chunk_id:
                chunk_name = chunk_name + str(parallel_dims_chunk_id[d]) + "."
            else:
                chunk_name = chunk_name + "0."
        chunk_name = chunk_name[:-1]

        if list(img_xds[data_variable_name].shape) != list(chunks):
            array = pad_array_with_nans(
                img_xds[data_variable_name].values,
                output_shape=chunks,
                dtype=dtype,
            )
        else:
            array = img_xds[data_variable_name].values

        write_binary_blob_to_disk(
            array,
            file_path=os.path.join(image_file, data_variable_name, chunk_name),
            compressor=compressor,
        )

        # z_chunk = zarr.open(
        #     os.path.join(image_file, data_variable_name, chunk_name),
        #     mode="a",
        #     shape=meta["shape"],
        #     chunks=meta["chunks"],
        #     dtype=meta["dtype"],
        #     compressor=compressor,
        # )

        # return z_chunk



import os
import time
from typing import Tuple, Union

import numpy as np
import xarray as xr

import toolviper.utils.logger as logger
from xradio.measurement_set._utils._msv2.msv4_sub_xdss import (
    rename_and_interpolate_to_time,
)
from xradio.measurement_set._utils._msv2.subtables import subt_rename_ids
from xradio.measurement_set._utils._msv2._tables.read import (
    convert_casacore_time_to_mjd,
    make_taql_where_between_min_max,
    load_generic_table,
)
from xradio._utils.list_and_array import cast_to_str
from xradio._utils.coord_math import (
    convert_to_si_units,
    add_position_offsets,
    wrap_to_pi,
)

from xradio._utils.list_and_array import (
    check_if_consistent,
    unique_1d,
    to_np_array,
)
from xradio._utils.schema import (
    casacore_to_msv4_measure_type,
    column_description_casacore_to_msv4_measure,
    convert_generic_xds_to_xradio_schema,
)


def create_field_and_source_xds(
    in_file: str,
    field_id: list,
    spectral_window_id: int,
    field_times: list,
    is_single_dish: bool,
    time_min_max: Tuple[np.float64, np.float64],
    ephemeris_interp_time: Union[xr.DataArray, None] = None,
) -> tuple[xr.Dataset, int]:
    """
    Create a field and source xarray dataset (xds) from the given input file, field ID, and spectral window ID.
    Data is extracted from the FIELD and SOURCE tables and if there is ephemeris data, it is also extracted.
    field_id will only be guaranteed to be a list of length 1 when partition_scheme does not include "FIELD_ID".

    Parameters:
    ----------
    in_file : str
        The path to the input file.
    field_id : list of int
        The field ids to select.
    spectral_window_id : int
        The ID of the spectral window.
    field_times: list
        Time data for field. It is the same as the time axis in the main MSv4 dataset and is used if more than one field is present.
    is_single_dish: bool
        whether the main xds has single-dish (SPECTRUM) data
    time_min_max : Tuple[np.float64, np.float46]
        Min / max times to constrain loading (usually to the time range relevant to an MSv4)
    ephemeris_interp_time : Union[xr.DataArray, None]
        Time axis to interpolate the ephemeris data vars to (usually main MSv4 time)

    Returns:
    -------
    field_and_source_xds : xr.Dataset
        The xarray dataset containing the field and source information.
    num_lines : int
        Sum of num_lines for all unique sources.
    """

    start_time = time.time()

    field_and_source_xds = xr.Dataset(attrs={"type": "field_and_source"})

    field_and_source_xds, ephemeris_path, ephemeris_table_name, source_id = (
        extract_field_info_and_check_ephemeris(
            field_and_source_xds, in_file, field_id, field_times, is_single_dish
        )
    )

    if field_and_source_xds.attrs["is_ephemeris"]:
        field_and_source_xds = extract_ephemeris_info(
            field_and_source_xds,
            ephemeris_path,
            ephemeris_table_name,
            is_single_dish,
            time_min_max,
            ephemeris_interp_time,
        )

    field_and_source_xds, num_lines = extract_source_info(
        field_and_source_xds, in_file, source_id, spectral_window_id
    )

    logger.debug(
        f"create_field_and_source_xds() execution time {time.time() - start_time:0.2f} s"
    )

    # Check if we can drop time axis. The phase centers are repeated.
    if field_times is not None:
        if is_single_dish:
            center_dv = "FIELD_REFERENCE_CENTER"
        else:
            center_dv = "FIELD_PHASE_CENTER"

        if np.unique(field_and_source_xds[center_dv], axis=0).shape[0] == 1:
            field_and_source_xds = field_and_source_xds.isel(time=0).drop_vars("time")

    return field_and_source_xds, source_id, num_lines


def extract_ephemeris_info(
    xds,
    path,
    table_name,
    is_single_dish,
    time_min_max: Tuple[np.float64, np.float64],
    interp_time: Union[xr.DataArray, None],
):
    """
    Extracts ephemeris information from the given path and table name and adds it to the xarray dataset.

    Parameters:
    ----------
    xds : xr.Dataset
        The xarray dataset to which the ephemeris information will be added.
    path : str
        The path to the input file.
    table_name : str
        The name of the ephemeris table.
    time_min_max : Tuple[np.float46, np.float64]
        Min / max times to constrain loading (usually to the time range relevant to an MSv4)
    interp_time : Union[xr.DataArray, None]
        Time axis to interpolate the data vars to (usually main MSv4 time)

    Returns:
    -------
    xds : xr.Dataset
        The xarray dataset with the added ephemeris information.
    """
    # The JPL-Horizons ephemeris table implementation in CASA does not follow the standard way of defining measures.
    # Consequently a lot of hardcoding is needed to extract the information.
    # https://casadocs.readthedocs.io/en/latest/notebooks/external-data.html

    # Only read data between the min and max times of the visibility data in the MSv4.
    min_max_mjd = (
        convert_casacore_time_to_mjd(time_min_max[0]),
        convert_casacore_time_to_mjd(time_min_max[1]),
    )
    taql_time_range = make_taql_where_between_min_max(
        min_max_mjd, path, table_name, "MJD"
    )
    ephemeris_xds = load_generic_table(
        path, table_name, timecols=["MJD"], taql_where=taql_time_range
    )

    assert (
        len(ephemeris_xds.ephemeris_id) == 1
    ), "Non standard ephemeris table. Only a single ephemeris is allowed per MSv4."
    ephemeris_xds = ephemeris_xds.isel(
        ephemeris_id=0
    )  # Collapse the ephemeris_id dimension.
    # Data varaibles  ['time', 'RA', 'DEC', 'Rho', 'RadVel', 'NP_ang', 'NP_dist', 'DiskLong', 'DiskLat', 'Sl_lon', 'Sl_lat', 'r', 'rdot', 'phang']

    # Get meta data.
    ephemeris_meta = ephemeris_xds.attrs["other"]["msv2"]["ctds_attrs"]
    ephemeris_column_description = ephemeris_xds.attrs["other"]["msv2"]["ctds_attrs"][
        "column_descriptions"
    ]

    assert (
        ephemeris_meta["obsloc"] == "GEOCENTRIC"
    ), "Only geocentric observer ephemeris are supported."

    if "posrefsys" in ephemeris_meta:
        # Note the phase center can be given as "J2000" or "J2000.0"
        ref_frame = (
            ephemeris_meta["posrefsys"]
            .replace("ICRF/", "", 1)
            .replace("J2000.0", "J2000", 1)
        )
        if ref_frame in casacore_to_msv4_measure_type["direction"].get("Ref_map", {}):
            ref_frame = casacore_to_msv4_measure_type["direction"]["Ref_map"][ref_frame]
        else:
            logger.debug(
                f"Unrecognized casacore direction reference frame found in posrefsys: {ref_frame}"
            )
        sky_coord_frame = ref_frame.lower()
    else:
        sky_coord_frame = "icrs"  # We will have to just assume this.

    # Find out witch keyword is used for units (UNIT/QuantumUnits)
    if "UNIT" in ephemeris_column_description["RA"]["keywords"]:
        unit_keyword = "UNIT"
    else:
        unit_keyword = "QuantumUnits"

    temp_xds = xr.Dataset()

    # Add mandatory data: OBSERVER_POSITION
    observer_position = [
        ephemeris_meta["GeoLong"],
        ephemeris_meta["GeoLat"],
        ephemeris_meta["GeoDist"],
    ]
    temp_xds["OBSERVER_POSITION"] = xr.DataArray(
        observer_position, dims=["ellipsoid_pos_label"]
    )
    temp_xds["OBSERVER_POSITION"].attrs.update(
        {
            "type": "location",
            "units": ["deg", "deg", "m"],
            "data": observer_position,
            "frame": "WGS84",
            "origin_object_name": "Earth",
            "coordinate_system": ephemeris_meta["obsloc"].lower(),
        }
    )  # I think the units are ['deg','deg','m'] and 'WGS84'.

    # Add (optional) data: SOURCE_LOCATION (POSITION / sky_pos_label)
    temp_xds["SOURCE_LOCATION"] = xr.DataArray(
        np.column_stack(
            (
                ephemeris_xds["RA"].data,
                ephemeris_xds["DEC"].data,
                ephemeris_xds["Rho"].data,
            )
        ),
        dims=["time_ephemeris", "sky_pos_label"],
    )
    # Have to use cast_to_str because the ephemeris table units are not consistently in a list or a string.
    sky_coord_units = [
        cast_to_str(ephemeris_column_description["RA"]["keywords"][unit_keyword]),
        cast_to_str(ephemeris_column_description["DEC"]["keywords"][unit_keyword]),
        cast_to_str(ephemeris_column_description["Rho"]["keywords"][unit_keyword]),
    ]
    temp_xds["SOURCE_LOCATION"].attrs.update(
        {"type": "sky_coord", "frame": sky_coord_frame, "units": sky_coord_units}
    )

    # Convert a few columns/variables that can be converted with standard
    # convert_generic_xds_to_xradio_schema().
    # Metadata has to be fixed manually. Alternatively, issues like
    # UNIT/QuantumUnits issue could be handled in convert_generic_xds_to_xradio_schema,
    # but for now preferring not to pollute that function.
    time_ephemeris_dim = ["time_ephemeris"]
    to_new_data_variables = {
        # mandatory: SOURCE_RADIAL_VELOCITY
        "RadVel": ["SOURCE_RADIAL_VELOCITY", time_ephemeris_dim],
        # optional: data NORTH_POLE_POSITION_ANGLE and NORTH_POLE_ANGULAR_DISTANCE
        "NP_ang": ["NORTH_POLE_POSITION_ANGLE", time_ephemeris_dim],
        "NP_dist": ["NORTH_POLE_ANGULAR_DISTANCE", time_ephemeris_dim],
        # optional: HELIOCENTRIC_RADIAL_VELOCITY
        "rdot": ["HELIOCENTRIC_RADIAL_VELOCITY", time_ephemeris_dim],
        # optional: OBSERVER_PHASE_ANGLE
        "phang": ["OBSERVER_PHASE_ANGLE", time_ephemeris_dim],
    }
    convert_generic_xds_to_xradio_schema(
        ephemeris_xds, temp_xds, to_new_data_variables, {}
    )

    # Adjust metadata:
    for generic_var_name, msv4_variable_def in to_new_data_variables.items():
        msv4_var_name = msv4_variable_def[0]
        if msv4_var_name in temp_xds:
            temp_xds[msv4_var_name].attrs.update(
                {
                    "type": "quantity",
                    "units": [
                        cast_to_str(
                            ephemeris_column_description[generic_var_name]["keywords"][
                                unit_keyword
                            ]
                        )
                    ],
                }
            )

    # Add optional data: SUB_OBSERVER_DIRECTION and SUB_SOLAR_POSITION
    if "DiskLong" in ephemeris_column_description:
        key_lon = "DiskLong"
        key_lat = "DiskLat"
    else:
        key_lon = "diskLong"
        key_lat = "diskLat"

    if key_lon in ephemeris_xds.data_vars:
        temp_xds["SUB_OBSERVER_DIRECTION"] = xr.DataArray(
            np.column_stack(
                (
                    ephemeris_xds[key_lon].data,
                    ephemeris_xds[key_lat].data,
                    np.zeros(ephemeris_xds[key_lon].shape),
                )
            ),
            dims=["time_ephemeris", "ellipsoid_pos_label"],
        )

        temp_xds["SUB_OBSERVER_DIRECTION"].attrs.update(
            {
                "type": "location",
                "frame": "Undefined",
                "origin_object_name": ephemeris_meta["NAME"],
                "coordinate_system": "planetodetic",
                "units": [
                    cast_to_str(
                        ephemeris_column_description[key_lon]["keywords"][unit_keyword]
                    ),
                    cast_to_str(
                        ephemeris_column_description[key_lat]["keywords"][unit_keyword]
                    ),
                    "m",
                ],
            }
        )

    if "SI_lon" in ephemeris_xds.data_vars:
        temp_xds["SUB_SOLAR_POSITION"] = xr.DataArray(
            np.column_stack(
                (
                    ephemeris_xds["SI_lon"].data,
                    ephemeris_xds["SI_lat"].data,
                    ephemeris_xds["r"].data,
                )
            ),
            dims=["time_ephemeris", "ellipsoid_pos_label"],
        )
        temp_xds["SUB_SOLAR_POSITION"].attrs.update(
            {
                "type": "location",
                "frame": "Undefined",
                "origin_object_name": "Sun",
                "coordinate_system": "planetodetic",
                "units": [
                    cast_to_str(
                        ephemeris_column_description["SI_lon"]["keywords"][unit_keyword]
                    ),
                    cast_to_str(
                        ephemeris_column_description["SI_lat"]["keywords"][unit_keyword]
                    ),
                    cast_to_str(
                        ephemeris_column_description["r"]["keywords"][unit_keyword]
                    ),
                ],
            }
        )

    # We are using the "time_ephemeris" label because it might not match the optional time axis of the source and field info. If ephemeris_interpolate=True then rename it to time.
    coords = {
        "ellipsoid_pos_label": ["lon", "lat", "dist"],
        "time_ephemeris": ephemeris_xds["time"].data,
        "sky_pos_label": ["ra", "dec", "dist"],
    }
    temp_xds = temp_xds.assign_coords(coords)
    time_coord_attrs = {
        "type": "time",
        "units": ["s"],
        "scale": "utc",
        "format": "unix",
    }
    temp_xds["time_ephemeris"].attrs.update(time_coord_attrs)

    # Convert to si units
    temp_xds = convert_to_si_units(temp_xds)

    # interpolate if ephemeris_interpolate/interp_time=True, and rename time_ephemeris=>time
    temp_xds = rename_and_interpolate_to_time(
        temp_xds, "time_ephemeris", interp_time, "field_and_source_xds"
    )

    xds = xr.merge([xds, temp_xds])

    # Add the SOURCE_LOCATION to the FIELD_PHASE_CENTER or FIELD_REFERENCE_CENTER. Ephemeris obs: When loaded from the MSv2 field table the FIELD_REFERENCE_CENTER or FIELD_PHASE_CENTER only contain an offset from the SOURCE_LOCATION.
    # We also need to add a distance dimension to the FIELD_PHASE_CENTER or FIELD_REFERENCE_CENTER to match the SOURCE_LOCATION.
    # FIELD_PHASE_CENTER is used for interferometer data and FIELD_REFERENCE_CENTER is used for single dish data.
    if is_single_dish:
        center_dv = "FIELD_REFERENCE_CENTER"
    else:
        center_dv = "FIELD_PHASE_CENTER"

    if "time" in xds[center_dv].coords:
        assert (
            interp_time is not None
        ), 'ephemeris_interpolate must be True if there is ephemeris data and multiple fields (this will occur if "FIELD_ID" is not in partition_scheme).'

        field_phase_center = wrap_to_pi(
            xds[center_dv].values + xds["SOURCE_LOCATION"][:, 0:2].values
        )
        field_phase_center = np.column_stack(
            (field_phase_center, np.zeros(xds[center_dv].values.shape[0]))
        )
        field_phase_center[:, -1] = (
            field_phase_center[:, -1] + xds["SOURCE_LOCATION"][:, -1].values
        )

        xds[center_dv] = xr.DataArray(
            field_phase_center,
            dims=[xds["SOURCE_LOCATION"].dims[0], "sky_pos_label"],
        )
    else:
        field_phase_center = (
            np.append(xds[center_dv].values, 0) + xds["SOURCE_LOCATION"].values
        )
        field_phase_center[:, 0:2] = wrap_to_pi(field_phase_center[:, 0:2])

        xds[center_dv] = xr.DataArray(
            field_phase_center,
            dims=[xds["SOURCE_LOCATION"].dims[0], "sky_pos_label"],
        )

    xds[center_dv].attrs.update(xds["SOURCE_LOCATION"].attrs)

    return xds


def make_line_dims_and_coords(
    source_xds: xr.Dataset, source_id: Union[int, np.ndarray], num_lines: int
) -> tuple[list, dict]:
    """
    Produces the dimensions and coordinates used in data variables related
    to line information (LINE_REST_FREQUENCY, LINE_SYSTEMIC_VELOCITY).

    In the dimensions, "time" is optional. To produce the points of the
    coordinates we need to look into the (optional) TRANSITION column or
    alternatively other columns (DIRECTION) to produce coordinates points of
    appropriate shape, given the "num_lines" "and source_id".

    Parameters:
    ----------
    source_xds: xr.Dataset
        generic source xarray dataset
    source_id: Union[int, np.ndarray]
        source_id of the dataset, when it is an array that indicates the
        presence of the "time" dimension
    num_line: int
        number of lines in the source dataset

    Returns:
    -------
    tuple : tuple[list, dict]
        The dimensions and coordinates to use with line data variables. The
        dimensions are produced as a list of dimension names, and the
        coordinates as a dict for xarray coords.
    """

    # Transition is an optional column and occasionally not populated
    if "TRANSITION" in source_xds.data_vars:
        transition_var_data = source_xds["TRANSITION"].data
    else:
        transition_var_data = np.zeros(source_xds["DIRECTION"].shape, dtype="str")

    # if TRANSITION is left empty (or otherwise incomplete), and num_lines > 1,
    # the data_vars expect a "num_lines" size in the last dimension
    vars_shape = transition_var_data.shape[:-1] + (np.max(num_lines),)
    if transition_var_data.shape == vars_shape:
        coords_lines_data = transition_var_data
    else:
        coords_lines_data = np.broadcast_to(
            transition_var_data, max(transition_var_data.shape, vars_shape)
        )

    line_label_data = np.arange(coords_lines_data.shape[-1]).astype(str)
    if len(source_id) == 1:
        line_coords = {
            "line_name": ("line_label", coords_lines_data),
            "line_label": line_label_data,
        }
        line_dims = ["line_label"]
    else:
        line_coords = {
            "line_name": (("time", "line_label"), coords_lines_data),
            "line_label": line_label_data,
        }
        line_dims = ["time", "line_label"]

    return line_dims, line_coords


def pad_missing_sources(
    source_xds: xr.Dataset, unique_source_ids: np.array
) -> xr.Dataset:
    """
    In some MSs there can be source IDs referenced from the field subtable which do not exist in
    the source table: https://github.com/casangi/xradio/issues/266

    This addresses the issue by padding/filling those IDs with "Unknown"/nan values. Produces a
    source_xds that, in addition to the information loaded for the non-missing source IDs, has
    padding for the IDs that are missing from the input MSv2 source table.
    This function does not need to do anything when unique_source_ids is a single value
    (partitioning by "FIELD_ID" or othwerwise single field/source)

    Parameters:
    ----------
    xds: xr.Dataset
        source dataset to fix/pad missing sources
    unique_source_ids: np.array
        IDs of the sources included in this partition

    Returns:
    -------
    filled_source_xds : xr.Dataset
        source dataset with padding in the originally missing sources
    """

    # Only fill gaps in multi-source xdss. If single source_id, no need to
    if len(unique_source_ids) <= 1:
        return source_xds

    missing_source_ids = [
        source_id
        for source_id in unique_source_ids
        if source_id not in source_xds.coords["SOURCE_ID"]
    ]

    # would like to use the new-ish xr.pad, but it creates issues with indices/coords and is
    # also not free of overheads, as it for example changes all numeric types to float64
    missing_source_xds = xr.full_like(source_xds.isel(SOURCE_ID=0), fill_value=np.nan)
    pad_str = "Unknown"
    pad_str_type = "<U9"
    for var in missing_source_xds.data_vars:
        if np.issubdtype(missing_source_xds.data_vars[var].dtype, np.str_):
            # Avoid truncation to length of previously loaded strings
            missing_source_xds[var] = missing_source_xds[var].astype(
                np.dtype(pad_str_type)
            )
            missing_source_xds[var] = pad_str

    concat_dim = "SOURCE_ID"
    xdss_to_concat = [source_xds]
    for missing_id in missing_source_ids:
        missing_source_xds[concat_dim] = missing_id
        xdss_to_concat.append(missing_source_xds)
    filled_source_xds = xr.concat(xdss_to_concat, concat_dim).sortby(concat_dim)

    return filled_source_xds


def extract_source_info(
    xds: xr.Dataset,
    path: str,
    source_id: Union[int, np.ndarray],
    spectral_window_id: int,
) -> tuple[xr.Dataset, int]:
    """
    Extracts source information from the given path and adds it to the xarray dataset.

    Parameters:
    ----------
    xds : xr.Dataset
        The xarray dataset to which the source information will be added.
    path : str
        The path to the input file.
    source_id : Union[int, np.ndarray]
        The ID of the source.
    spectral_window_id : int
        The ID of the spectral window.

    Returns:
    -------
    xds : xr.Dataset
        The xarray dataset with the added source information.
    num_lines : int
        Sum of num_lines for all unique sources extracted.
    """
    coords = {}
    is_ephemeris = xds.attrs[
        "is_ephemeris"
    ]  # If ephemeris data is present we ignore the SOURCE_DIRECTION in the source table.

    if all(source_id == -1):
        logger.warning(
            f"Source_id is -1. No source information will be included in the field_and_source_xds."
        )
        xds = xds.assign_coords(
            {"source_name": "Unknown"}
        )  # Need to add this for ps.summary() to work.
        return xds, 0

    if not os.path.isdir(os.path.join(path, "SOURCE")):
        logger.warning(
            f"Could not find SOURCE table for source_id {source_id}. Source information will not be included in the field_and_source_xds."
        )
        xds = xds.assign_coords({"source_name": "Unknown"})
        return xds, 0

    unique_source_id = unique_1d(source_id)
    taql_where = f"where (SOURCE_ID IN [{','.join(map(str, unique_source_id))}]) AND (SPECTRAL_WINDOW_ID = {spectral_window_id})"

    source_xds = load_generic_table(
        path,
        "SOURCE",
        ignore=["SOURCE_MODEL"],  # Trying to read SOURCE_MODEL causes an error.
        taql_where=taql_where,
    )

    if len(source_xds.data_vars) == 0:  # The source xds is empty.
        logger.warning(
            f"SOURCE table empty for (unique) source_id {unique_source_id} and spectral_window_id {spectral_window_id}."
        )
        xds = xds.assign_coords(
            {"source_name": "Unknown"}
        )  # Need to add this for ps.summary() to work.
        return xds, 0

    assert (
        len(source_xds.SPECTRAL_WINDOW_ID) == 1
    ), "Can only process source table with a single spectral_window_id for a given MSv4 partition."

    # This source table time is not the same as the time in the field_and_source_xds that is derived from the main MSv4 time axis.
    # The source_id maps to the time axis in the field_and_source_xds. That is why "if len(source_id) == 1" is used to check if there should be a time axis.
    # assert len(source_xds.TIME) <= len(
    #     unique_source_id
    # ), "Can only process source table with a single time entry for a source_id and spectral_window_id."
    if len(source_xds.TIME) > len(unique_source_id):
        logger.warning(
            f"Source table has more than one time entry for a source_id and spectral_window_id. This is not currently supported. Only the first time entry will be used."
        )
        source_xds = source_xds.drop_duplicates("SOURCE_ID", keep="first")

    source_xds = source_xds.isel(TIME=0, SPECTRAL_WINDOW_ID=0, drop=True)
    source_column_description = source_xds.attrs["other"]["msv2"]["ctds_attrs"][
        "column_descriptions"
    ]

    source_xds = pad_missing_sources(source_xds, unique_source_id)

    # Get source name (the time axis is optional and will probably be required if the partition scheme does not include 'FIELD_ID' or 'SOURCE_ID'.).
    # Note again that this optional time axis has nothing to do with the original time axis in the source table that we drop.
    if len(source_id) == 1:
        source_xds = source_xds.sel(SOURCE_ID=source_id[0])
        coords["source_name"] = (
            source_xds["NAME"].values.item() + "_" + str(source_id[0])
        )
        direction_dims = ["sky_dir_label"]
        # coords["source_id"] = source_id[0]
    else:
        source_xds = source_xds.sel(SOURCE_ID=source_id)
        coords["source_name"] = (
            "time",
            np.char.add(
                source_xds["NAME"].data, np.char.add("_", source_id.astype(str))
            ),
        )
        direction_dims = ["time", "sky_dir_label"]
        # coords["source_id"] = ("time", source_id)

    # If ephemeris data is present we ignore the SOURCE_DIRECTION.
    if not is_ephemeris:
        direction_msv2_col = "DIRECTION"
        msv4_measure = column_description_casacore_to_msv4_measure(
            source_column_description[direction_msv2_col]
        )

        msv2_direction_dims = source_xds[direction_msv2_col].dims
        if (
            len(msv2_direction_dims) == len(direction_dims) + 1
            and "dim_1" in msv2_direction_dims
            and "dim_2" in msv2_direction_dims
        ):
            # CASA simulator produces transposed direction values, adding an
            # unexpected dimension. Drop it (https://github.com/casangi/xradio/issues/#196)
            direction_var = source_xds[direction_msv2_col].isel(dim_1=0, drop=True)
        else:
            direction_var = source_xds[direction_msv2_col]

        # SOURCE_LOCATION (DIRECTION / sky_dir_label)
        location_msv4_measure = column_description_casacore_to_msv4_measure(
            source_column_description[direction_msv2_col]
        )
        xds["SOURCE_LOCATION"] = xr.DataArray(
            direction_var.data, dims=direction_dims, attrs=location_msv4_measure
        )

    # Do we have line data:
    if source_xds["NUM_LINES"].data.ndim == 0:
        num_lines = np.array([source_xds["NUM_LINES"].data.item()])
    else:
        num_lines = source_xds["NUM_LINES"].data

    if any(num_lines > 0):
        line_dims, line_coords = make_line_dims_and_coords(
            source_xds, source_id, num_lines
        )
        xds = xds.assign_coords(line_coords)

        to_new_data_variables = {
            "REST_FREQUENCY": ["LINE_REST_FREQUENCY", line_dims],
            "SYSVEL": ["LINE_SYSTEMIC_VELOCITY", line_dims],
        }
        to_new_coords = {
            "TIME": ["time", ["time"]],
        }
        convert_generic_xds_to_xradio_schema(
            source_xds, xds, to_new_data_variables, to_new_coords
        )

    # Need to add doppler info if present. Add check.
    try:
        doppler_xds = load_generic_table(
            path,
            "DOPPLER",
        )
        assert (
            False
        ), "Doppler table present. Please open an issue on https://github.com/casangi/xradio/issues so that we can add support for this."
    except:
        pass

    xds = xds.assign_coords(coords)

    _, unique_source_ids_indices = np.unique(source_xds.SOURCE_ID, return_index=True)

    return xds, np.sum(num_lines[unique_source_ids_indices])


def make_field_dims_and_coords(
    field_xds: xr.Dataset, field_id: Union[int, np.ndarray], field_times: list
) -> tuple[list, dict]:
    """
    Produces the dimensions and coordinates used in the field data variables
    extracted from the MSv2 FIELD subtable (FIELD_PHASE_CENTER/
    FIELD_REFERENCE_CENTER).

    Parameters:
    ----------
    field_xds: xr.Dataset
        generic field xarray dataset
    field_id: Union[int, np.ndarray]
        field_id of the dataset
    field_times:
        Unique times for the dataset (when not partitioning by FIELD_ID)

    Returns:
    -------
    tuple : tuple[list, dict]
        The dimensions and coordinates to use with field data variables. The
        dimensions are produced as a list of dimension names, and the
        coordinates as a dict for xarray coords.
    """

    coords = {"sky_dir_label": ["ra", "dec"]}

    # field_times is the same as the time axis in the main MSv4 dataset and is used if more than one field is present.
    if field_times is not None:
        coords["time"] = field_times
        dims = ["time", "sky_dir_label"]
        coords["field_name"] = (
            "time",
            np.char.add(field_xds["NAME"].data, np.char.add("_", field_id.astype(str))),
        )
        # coords["field_id"] = ("time", field_id)
    else:
        coords["field_name"] = field_xds["NAME"].values.item() + "_" + str(field_id)
        # coords["field_id"] = field_id
        dims = ["sky_dir_label"]

    return dims, coords


def extract_field_info_and_check_ephemeris(
    field_and_source_xds: xr.Dataset,
    in_file: str,
    field_id: Union[int, np.ndarray],
    field_times: list,
    is_single_dish: bool,
):
    """
    Create field information and check for ephemeris in the FIELD table folder.

    Parameters:
    ----------
    field_and_source_xds : xr.Dataset
        The xarray dataset to which the field and source information will be added.
    in_file : str
        The path to the input file.
    field_id : Union[int, np.ndarray]
        The ID of the field.
    field_times: list
        Time of the MSv4
    is_single_dish: bool
        Whether to extract single dish (FIELD_REFERENCE_CENTER) info

    Returns:
    -------
    field_and_source_xds : xr.Dataset
        The xarray dataset with the added field and source information.
    ephemeris_path : str
        The path to the ephemeris table.
    ephemeris_table_name : str
        The name of the ephemeris table.
    """

    unique_field_id = unique_1d(
        field_id
    )  # field_ids can be repeated so that the time mapping is correct if there are multiple fields. The load_generic_table required unique field_ids.
    taql_where = f"where (ROWID() IN [{','.join(map(str, unique_field_id))}])"
    field_xds = load_generic_table(
        in_file,
        "FIELD",
        rename_ids=subt_rename_ids["FIELD"],
        taql_where=taql_where,
    )

    assert (
        len(field_xds.poly_id) == 1
    ), "Polynomial field positions not supported. Please open an issue on https://github.com/casangi/xradio/issues so that we can add support for this."
    field_xds = field_xds.isel(poly_id=0, drop=True)
    # field_xds = field_xds.assign_coords({'field_id':field_xds['field_id'].data})
    field_xds = field_xds.assign_coords({"field_id": unique_field_id})
    field_xds = field_xds.sel(
        field_id=field_id, drop=False
    )  # Make sure field_id match up with time axis (duplicate fields are allowed).
    source_id = to_np_array(field_xds.SOURCE_ID.values)

    ephemeris_table_name = None
    ephemeris_path = None
    is_ephemeris = False
    field_and_source_xds.attrs["is_ephemeris"] = (
        False  # If we find a path to the ephemeris table we will set this to True.
    )

    # Need to check if ephemeris_id is present and if ephemeris table is present.
    if "EPHEMERIS_ID" in field_xds:
        # Note: this assumes partition_scheme includes "FIELD_ID"
        ephemeris_id = check_if_consistent(field_xds.EPHEMERIS_ID, "EPHEMERIS_ID")

        if ephemeris_id > -1:
            files = os.listdir(os.path.join(in_file, "FIELD"))
            ephemeris_table_name_start = "EPHEM" + str(ephemeris_id)

            ephemeris_name_table_index = [
                i for i in range(len(files)) if ephemeris_table_name_start in files[i]
            ]

            assert len(ephemeris_name_table_index) == 1, (
                "More than one ephemeris table which starts with "
                + ephemeris_table_name_start
            )

            if len(ephemeris_name_table_index) > 0:  # Are there any ephemeris tables.
                is_ephemeris = True
                e_index = ephemeris_name_table_index[0]
                ephemeris_path = os.path.join(in_file, "FIELD")
                ephemeris_table_name = files[e_index]
                field_and_source_xds.attrs["is_ephemeris"] = True
            else:
                logger.warning(
                    f"Could not find ephemeris table for field_id {field_id}. Ephemeris information will not be included in the field_and_source_xds."
                )

    dims, coords = make_field_dims_and_coords(field_xds, field_id, field_times)

    if is_single_dish:
        field_data_variables = {
            "REFERENCE_DIR": "FIELD_REFERENCE_CENTER",
        }
    else:
        field_data_variables = {
            # "DELAY_DIR": "FIELD_DELAY_CENTER",
            "PHASE_DIR": "FIELD_PHASE_CENTER",
            # "REFERENCE_DIR": "FIELD_REFERENCE_CENTER",
        }

    field_column_description = field_xds.attrs["other"]["msv2"]["ctds_attrs"][
        "column_descriptions"
    ]

    for generic_name, msv4_name in field_data_variables.items():

        delay_dir_ref_col = "DelayDir_Ref"
        if field_xds.get(delay_dir_ref_col) is None:
            delaydir_ref = None
        else:
            delaydir_ref = check_if_consistent(
                field_xds.get(delay_dir_ref_col), delay_dir_ref_col
            )

        msv4_measure = column_description_casacore_to_msv4_measure(
            field_column_description[generic_name], ref_code=delaydir_ref
        )

        field_and_source_xds[msv4_name] = xr.DataArray.from_dict(
            {
                "dims": dims,
                "data": list(field_xds[generic_name].data),
                "attrs": msv4_measure,
            }
        )

        field_measures_type = "sky_coord"
        field_and_source_xds[msv4_name].attrs["type"] = field_measures_type

    field_and_source_xds = field_and_source_xds.assign_coords(coords)
    if "time" in field_and_source_xds:
        time_column_description = field_xds.attrs["other"]["msv2"]["ctds_attrs"][
            "column_descriptions"
        ]["TIME"]
        time_msv4_measure = column_description_casacore_to_msv4_measure(
            time_column_description
        )
        field_and_source_xds.coords["time"].attrs.update(time_msv4_measure)

    return field_and_source_xds, ephemeris_path, ephemeris_table_name, source_id



import numcodecs, os, time
from pathlib import Path
from typing import Dict, Union

import zarr
import toolviper.utils.logger as logger

from ._utils.cds import CASAVisSet
from ._zarr.read import read_part_keys, read_partitions, read_subtables
from ._zarr.write import write_metainfo, write_part_keys, write_partitions


def is_zarr_cor(inpath: str) -> bool:
    """
    Check if a given path has a visibilities dataset in Zarr format

    Parameters
    ----------
    inpath : str
        path to a (possibly) Zarr vis dataset

    Returns
    -------
    bool
        whether zarr.open can open this path
    """
    try:
        with zarr.open(Path(inpath, "partition_keys"), mode="r"):
            logger.debug(f"{inpath} can be opened as Zarr format data")
            return True
    except zarr.errors.PathNotFoundError:
        return False


def read_cor(
    inpath: str,
    subtables: bool = True,
    asdm_subtables: bool = False,
) -> CASAVisSet:
    """
    Read a CASAVisSet stored in zarr format.

    Parameters
    ----------
    inpath : str
        Input Zarr path
    subtables : bool (Default value = True)
        Also read and (metainformation) subtables along with main visibilities data.
    asdm_subtables : bool (Default value = False)
        Also read extension subtables named "ASDM_*"

    Returns
    -------
    CASAVisSet
        Main xarray dataset of datasets for this visibility dataset
    """
    inpath = os.path.expanduser(inpath)
    if not os.path.isdir(inpath):
        raise ValueError(f"invalid input filename to read_vis {inpath}")

    logger.info(f"Reading {inpath} as visibilities dataset stored in Zarr format")

    all_start = time.time()

    metainfo = {}
    if subtables:
        metainfo = read_subtables(inpath, asdm_subtables)

    part_keys = read_part_keys(inpath)
    partitions = read_partitions(inpath, part_keys)

    all_time = time.time() - all_start
    logger.info(f"Time to read dataset from_zarr {inpath}: {all_time}")

    vers = "version-WIP"
    descr_add = "read_vis from zarr"
    cds = CASAVisSet(
        metainfo=metainfo,
        partitions=dict.fromkeys(part_keys, partitions),
        descr=f"CASA vis set produced by xradio {vers}/{descr_add}",
    )

    return cds


def write_cor(
    cds: CASAVisSet,
    outpath: str,
    chunks_on_disk: Union[Dict, None] = None,
    compressor: Union[numcodecs.abc.Codec, None] = None,
) -> None:
    """
    Write CASA vis dataset to zarr format on disk. When
    chunks_on_disk is not specified the chunking in the input dataset
    is used. When chunks_on_disk is specified that dataset is saved
    using that chunking.

    Parameters
    ----------
    cds : CASAVisSet
        CASA visibilities dataset to write to disk
    outpath : str
        output path, generally ends in .zarr
    chunks_on_disk : Union[Dict, None] = None (Default value = None)
        a dictionary with the chunk size that will
        be used when writing to disk. For example {'time': 20, 'chan': 6}.
        If chunks_on_disk is not specified the chunking of dataset will
        be used.
    compressor : Union[numcodecs.abc.Codec, None] (Default value = None)
        the blosc compressor to use when saving the
        converted data to disk using zarr. If None the zstd compression
        algorithm used with compression level 2.

    Returns
    -------
    """

    if compressor is None:
        compressor = numcodecs.Blosc(cname="zstd", clevel=2, shuffle=0)

    if os.path.lexists(outpath):
        raise ValueError(f"output vis.zarr path ({outpath}) already exists")

    all_start = time.time()

    write_part_keys(cds.partitions, outpath, compressor)

    write_metainfo(outpath, cds.metainfo, chunks_on_disk, compressor)

    write_partitions(outpath, cds.partitions, chunks_on_disk, compressor)

    all_time = time.time() - all_start
    logger.info(f"Time to prepare and save dataset to_zarr {outpath}: {all_time}")



import pandas as pd
from xradio._utils.list_and_array import to_list
import numbers
import numpy as np
import toolviper.utils.logger as logger
import xarray as xr


class ProcessingSet(dict):
    """
    A dictionary subclass representing a Processing Set (PS) containing Measurement Sets v4 (MS).

    This class extends the built-in `dict` class to provide additional methods for
    manipulating and selecting subsets of the Processing Set. It includes functionality
    for summarizing metadata, selecting subsets based on various criteria, and
    exporting the data to storage formats.

    Parameters
    ----------
    *args : dict, optional
        Variable length argument list passed to the base `dict` class.
    **kwargs : dict, optional
        Arbitrary keyword arguments passed to the base `dict` class.
    """

    def __init__(self, *args, **kwargs):
        """
        Initialize the ProcessingSet instance.

        Parameters
        ----------
        *args : dict, optional
            Variable length argument list passed to the base `dict` class.
        **kwargs : dict, optional
            Arbitrary keyword arguments passed to the base `dict` class.
        """
        super().__init__(*args, **kwargs)
        self.meta = {"summary": {}}

    def summary(self, data_group="base"):
        """
        Generate and retrieve a summary of the Processing Set.

        The summary includes information such as the names of the Measurement Sets,
        their intents, polarizations, spectral window names, field names, source names,
        field coordinates, start frequencies, and end frequencies.

        Parameters
        ----------
        data_group : str, optional
            The data group to summarize. Default is "base".

        Returns
        -------
        pandas.DataFrame
            A DataFrame containing the summary information of the specified data group.
        """

        if data_group in self.meta["summary"]:
            return self.meta["summary"][data_group]
        else:
            self.meta["summary"][data_group] = self._summary(data_group).sort_values(
                by=["name"], ascending=True
            )
            return self.meta["summary"][data_group]

    def get_ps_max_dims(self):
        """
        Determine the maximum dimensions across all Measurement Sets in the Processing Set.

        This method examines each Measurement Set's dimensions and computes the maximum
        size for each dimension across the entire Processing Set.

        For example, if the Processing Set contains two MSs with dimensions (50, 20, 30) and (10, 30, 40),
        the maximum dimensions will be (50, 30, 40).

        Returns
        -------
        dict
            A dictionary containing the maximum dimensions of the Processing Set, with dimension names as keys
            and their maximum sizes as values.
        """
        if "max_dims" in self.meta:
            return self.meta["max_dims"]
        else:
            self.meta["max_dims"] = self._get_ps_max_dims()
            return self.meta["max_dims"]

    def get_ps_freq_axis(self):
        """
        Combine the frequency axes of all Measurement Sets in the Processing Set.

        This method aggregates the frequency information from each Measurement Set to create
        a unified frequency axis for the entire Processing Set.

        Returns
        -------
        xarray.DataArray
            The combined frequency axis of the Processing Set.
        """
        if "freq_axis" in self.meta:
            return self.meta["freq_axis"]
        else:
            self.meta["freq_axis"] = self._get_ps_freq_axis()
            return self.meta["freq_axis"]

    def _summary(self, data_group="base"):
        summary_data = {
            "name": [],
            "intents": [],
            "shape": [],
            "polarization": [],
            "scan_number": [],
            "spw_name": [],
            "field_name": [],
            "source_name": [],
            "line_name": [],
            "field_coords": [],
            "start_frequency": [],
            "end_frequency": [],
        }
        from astropy.coordinates import SkyCoord
        import astropy.units as u

        for key, value in self.items():
            summary_data["name"].append(key)
            summary_data["intents"].append(value.attrs["partition_info"]["intents"])
            summary_data["spw_name"].append(
                value.attrs["partition_info"]["spectral_window_name"]
            )
            summary_data["polarization"].append(value.polarization.values)
            summary_data["scan_number"].append(
                value.attrs["partition_info"]["scan_number"]
            )
            data_name = value.attrs["data_groups"][data_group]["correlated_data"]

            if "VISIBILITY" in data_name:
                center_name = "FIELD_PHASE_CENTER"

            if "SPECTRUM" in data_name:
                center_name = "FIELD_REFERENCE_CENTER"

            summary_data["shape"].append(value[data_name].shape)

            summary_data["field_name"].append(
                value.attrs["partition_info"]["field_name"]
            )
            summary_data["source_name"].append(
                value.attrs["partition_info"]["source_name"]
            )

            summary_data["line_name"].append(value.attrs["partition_info"]["line_name"])

            summary_data["start_frequency"].append(
                to_list(value["frequency"].values)[0]
            )
            summary_data["end_frequency"].append(to_list(value["frequency"].values)[-1])

            if value[data_name].attrs["field_and_source_xds"].is_ephemeris:
                summary_data["field_coords"].append("Ephemeris")
            elif (
                "time"
                in value[data_name].attrs["field_and_source_xds"][center_name].coords
            ):
                summary_data["field_coords"].append("Multi-Phase-Center")
            else:
                ra_dec_rad = (
                    value[data_name].attrs["field_and_source_xds"][center_name].values
                )
                frame = (
                    value[data_name]
                    .attrs["field_and_source_xds"][center_name]
                    .attrs["frame"]
                    .lower()
                )

                coord = SkyCoord(
                    ra=ra_dec_rad[0] * u.rad, dec=ra_dec_rad[1] * u.rad, frame=frame
                )

                summary_data["field_coords"].append(
                    [
                        frame,
                        coord.ra.to_string(unit=u.hour, precision=2),
                        coord.dec.to_string(unit=u.deg, precision=2),
                    ]
                )

        summary_df = pd.DataFrame(summary_data)
        return summary_df

    def _get_ps_freq_axis(self):

        spw_ids = []
        freq_axis_list = []
        frame = self.get(0).frequency.attrs["observer"]
        for ms_xds in self.values():
            assert (
                frame == ms_xds.frequency.attrs["observer"]
            ), "Frequency reference frame not consistent in Processing Set."
            if ms_xds.frequency.attrs["spectral_window_id"] not in spw_ids:
                spw_ids.append(ms_xds.frequency.attrs["spectral_window_id"])
                freq_axis_list.append(ms_xds.frequency)

        freq_axis = xr.concat(freq_axis_list, dim="frequency").sortby("frequency")
        return freq_axis

    def _get_ps_max_dims(self):
        max_dims = None
        for ms_xds in self.values():
            if max_dims is None:
                max_dims = dict(ms_xds.sizes)
            else:
                for dim_name, size in ms_xds.sizes.items():
                    if dim_name in max_dims:
                        if max_dims[dim_name] < size:
                            max_dims[dim_name] = size
                    else:
                        max_dims[dim_name] = size
        return max_dims

    def get(self, id):
        return self[list(self.keys())[id]]

    def sel(self, string_exact_match: bool = True, query: str = None, **kwargs):
        """
        Select a subset of the Processing Set based on specified criteria.

        This method allows filtering the Processing Set by matching column names and values
        or by applying a Pandas query string. The selection criteria can target various
        attributes of the Measurement Sets such as intents, polarization, spectral window names, etc.

        Note
        ----
        This selection does not modify the actual data within the Measurement Sets. For example, if
        a Measurement Set has `field_name=['field_0','field_10','field_08']` and `ps.sel(field_name='field_0')`
        is invoked, the resulting subset will still contain the original list `['field_0','field_10','field_08']`.

        Parameters
        ----------
        string_exact_match : bool, optional
            If `True`, string matching will require exact matches for string and string list columns.
            If `False`, partial matches are allowed. Default is `True`.
        query : str, optional
            A Pandas query string to apply additional filtering. Default is `None`.
        **kwargs : dict
            Keyword arguments representing column names and their corresponding values to filter the Processing Set.

        Returns
        -------
        ProcessingSet
            A new `ProcessingSet` instance containing only the Measurement Sets that match the selection criteria.

        Examples
        --------
        >>> # Select all MSs with intents 'OBSERVE_TARGET#ON_SOURCE' and polarization 'RR' or 'LL'
        >>> selected_ps = ps.sel(intents='OBSERVE_TARGET#ON_SOURCE', polarization=['RR', 'LL'])

        >>> # Select all MSs with start_frequency greater than 100 GHz and less than 200 GHz
        >>> selected_ps = ps.sel(query='start_frequency > 100e9 AND end_frequency < 200e9')
        """
        import numpy as np

        def select_rows(df, col, sel_vals, string_exact_match):
            def check_selection(row_val):
                row_val = to_list(
                    row_val
                )  # make sure that it is a list so that we can iterate over it.

                for rw in row_val:
                    for s in sel_vals:
                        if string_exact_match:
                            if rw == s:
                                return True
                        else:
                            if s in rw:
                                return True
                return False

            return df[df[col].apply(check_selection)]

        summary_table = self.summary()
        for key, value in kwargs.items():
            value = to_list(value)  # make sure value is a list.

            if len(value) == 1 and isinstance(value[0], slice):
                summary_table = summary_table[
                    summary_table[key].between(value[0].start, value[0].stop)
                ]
            else:
                summary_table = select_rows(
                    summary_table, key, value, string_exact_match
                )

        if query is not None:
            summary_table = summary_table.query(query)

        sub_ps = ProcessingSet()
        for key, val in self.items():
            if key in summary_table["name"].values:
                sub_ps[key] = val

        return sub_ps

    def ms_sel(self, **kwargs):
        """
        Select a subset of the Processing Set by applying the `xarray.Dataset.sel` method to each Measurement Set.

        This method allows for selection based on label-based indexing for each dimension of the datasets.

        Parameters
        ----------
        **kwargs : dict
            Keyword arguments representing dimension names and the labels to select along those dimensions.
            These are passed directly to the `xarray.Dataset.sel <https://docs.xarray.dev/en/latest/generated/xarray.Dataset.sel.html>`__ method.

        Returns
        -------
        ProcessingSet
            A new `ProcessingSet` instance containing the selected subsets of each Measurement Set.
        """
        sub_ps = ProcessingSet()
        for key, val in self.items():
            sub_ps[key] = val.sel(kwargs)
        return sub_ps

    def ms_isel(self, **kwargs):
        """
        Select a subset of the Processing Set by applying the `isel` method to each Measurement Set.

        This method allows for selection based on integer-based indexing for each dimension of the datasets.

        Parameters
        ----------
        **kwargs : dict
            Keyword arguments representing dimension names and the integer indices to select along those dimensions.
            These are passed directly to the `xarray.Dataset.isel <https://docs.xarray.dev/en/latest/generated/xarray.Dataset.isel.html>`__ method.

        Returns
        -------
        ProcessingSet
            A new `ProcessingSet` instance containing the selected subsets of each Measurement Set.
        """
        sub_ps = ProcessingSet()
        for key, val in self.items():
            sub_ps[key] = val.isel(kwargs)
        return sub_ps

    def to_store(self, store, **kwargs):
        """
        Write the Processing Set to a Zarr store.

        This method serializes each Measurement Set within the Processing Set to a separate Zarr group
        within the specified store directory. Note that writing to cloud storage is not supported yet.

        Parameters
        ----------
        store : str
            The filesystem path to the Zarr store directory where the data will be saved.
        **kwargs : dict, optional
            Additional keyword arguments to be passed to the `xarray.Dataset.to_zarr` method.
            Refer to the `xarray.Dataset.to_zarr <https://docs.xarray.dev/en/latest/generated/xarray.Dataset.to_zarr.html>`__
            for available options.

        Returns
        -------
        None

        Raises
        ------
        OSError
            If the specified store path is invalid or not writable.

        Examples
        --------
        >>> # Save the Processing Set to a local Zarr store
        >>> ps.to_store('/path/to/zarr_store')
        """
        import os

        for key, value in self.items():
            value.to_store(os.path.join(store, key), **kwargs)

    def get_combined_field_and_source_xds(self, data_group="base"):
        """
        Combine the `field_and_source_xds` datasets from all Measurement Sets into a single dataset.

        The combined `xarray.Dataset` will have a new dimension 'field_name', consolidating data from
        each Measurement Set. Ephemeris data is handled separately.

        Parameters
        ----------
        data_group : str, optional
            The data group to process. Default is "base".

        Returns
        -------
        tuple of xarray.Dataset
            A tuple containing two `xarray.Dataset` objects:
            - combined_field_and_source_xds: Combined dataset for standard fields.
            - combined_ephemeris_field_and_source_xds: Combined dataset for ephemeris fields.

        Raises
        ------
        ValueError
            If the `field_and_source_xds` attribute is missing or improperly formatted in any Measurement Set.
        """
        df = self.summary(data_group)

        combined_field_and_source_xds = xr.Dataset()
        combined_ephemeris_field_and_source_xds = xr.Dataset()
        for ms_name, ms_xds in self.items():

            correlated_data_name = ms_xds.attrs["data_groups"][data_group][
                "correlated_data"
            ]

            field_and_source_xds = (
                ms_xds[correlated_data_name]
                .attrs["field_and_source_xds"]
                .copy(deep=True)
            )

            if (
                "line_name" in field_and_source_xds.coords
            ):  # Not including line info since it is a function of spw.
                field_and_source_xds = field_and_source_xds.drop_vars(
                    ["LINE_REST_FREQUENCY", "LINE_SYSTEMIC_VELOCITY"], errors="ignore"
                )
                del field_and_source_xds["line_name"]
                del field_and_source_xds["line_label"]

            if "time" in field_and_source_xds.coords:
                if "time" not in field_and_source_xds.field_name.dims:
                    field_names = np.array(
                        [field_and_source_xds.field_name.values.item()]
                        * len(field_and_source_xds.time.values)
                    )
                    source_names = np.array(
                        [field_and_source_xds.source_name.values.item()]
                        * len(field_and_source_xds.time.values)
                    )
                    del field_and_source_xds["field_name"]
                    del field_and_source_xds["source_name"]
                    field_and_source_xds = field_and_source_xds.assign_coords(
                        field_name=("time", field_names)
                    )
                    field_and_source_xds = field_and_source_xds.assign_coords(
                        source_name=("time", source_names)
                    )
                field_and_source_xds = field_and_source_xds.swap_dims(
                    {"time": "field_name"}
                )
                del field_and_source_xds["time"]
            elif "time_ephemeris" in field_and_source_xds.coords:
                if "time_ephemeris" not in field_and_source_xds.field_name.dims:
                    field_names = np.array(
                        [field_and_source_xds.field_name.values.item()]
                        * len(field_and_source_xds.time_ephemeris.values)
                    )
                    source_names = np.array(
                        [field_and_source_xds.source_name.values.item()]
                        * len(field_and_source_xds.time_ephemeris.values)
                    )
                    del field_and_source_xds["field_name"]
                    del field_and_source_xds["source_name"]
                    field_and_source_xds = field_and_source_xds.assign_coords(
                        field_name=("time_ephemeris", field_names)
                    )
                    field_and_source_xds = field_and_source_xds.assign_coords(
                        source_name=("time_ephemeris", source_names)
                    )
                field_and_source_xds = field_and_source_xds.swap_dims(
                    {"time_ephemeris": "field_name"}
                )
                del field_and_source_xds["time_ephemeris"]
            else:
                for dv_names in field_and_source_xds.data_vars:
                    if "field_name" not in field_and_source_xds[dv_names].dims:
                        field_and_source_xds[dv_names] = field_and_source_xds[
                            dv_names
                        ].expand_dims("field_name")

            if field_and_source_xds.is_ephemeris:
                if len(combined_ephemeris_field_and_source_xds.data_vars) == 0:
                    combined_ephemeris_field_and_source_xds = field_and_source_xds
                else:
                    combined_ephemeris_field_and_source_xds = xr.concat(
                        [combined_ephemeris_field_and_source_xds, field_and_source_xds],
                        dim="field_name",
                    )
            else:
                if len(combined_field_and_source_xds.data_vars) == 0:
                    combined_field_and_source_xds = field_and_source_xds
                else:
                    combined_field_and_source_xds = xr.concat(
                        [combined_field_and_source_xds, field_and_source_xds],
                        dim="field_name",
                    )

        if (len(combined_field_and_source_xds.data_vars) > 0) and (
            "FIELD_PHASE_CENTER" in combined_field_and_source_xds
        ):
            combined_field_and_source_xds = (
                combined_field_and_source_xds.drop_duplicates("field_name")
            )

            combined_field_and_source_xds["MEAN_PHASE_CENTER"] = (
                combined_field_and_source_xds["FIELD_PHASE_CENTER"].mean(
                    dim=["field_name"]
                )
            )

            ra1 = (
                combined_field_and_source_xds["FIELD_PHASE_CENTER"]
                .sel(sky_dir_label="ra")
                .values
            )
            dec1 = (
                combined_field_and_source_xds["FIELD_PHASE_CENTER"]
                .sel(sky_dir_label="dec")
                .values
            )
            ra2 = (
                combined_field_and_source_xds["MEAN_PHASE_CENTER"]
                .sel(sky_dir_label="ra")
                .values
            )
            dec2 = (
                combined_field_and_source_xds["MEAN_PHASE_CENTER"]
                .sel(sky_dir_label="dec")
                .values
            )

            from xradio._utils.coord_math import haversine

            distance = haversine(ra1, dec1, ra2, dec2)
            min_index = distance.argmin()

            combined_field_and_source_xds.attrs["center_field_name"] = (
                combined_field_and_source_xds.field_name[min_index].values
            )

        if (len(combined_ephemeris_field_and_source_xds.data_vars) > 0) and (
            "FIELD_PHASE_CENTER" in combined_ephemeris_field_and_source_xds
        ):
            combined_ephemeris_field_and_source_xds = (
                combined_ephemeris_field_and_source_xds.drop_duplicates("field_name")
            )

            from xradio._utils.coord_math import wrap_to_pi

            offset = (
                combined_ephemeris_field_and_source_xds["FIELD_PHASE_CENTER"]
                - combined_ephemeris_field_and_source_xds["SOURCE_LOCATION"]
            )
            combined_ephemeris_field_and_source_xds["FIELD_OFFSET"] = xr.DataArray(
                wrap_to_pi(offset.sel(sky_pos_label=["ra", "dec"])).values,
                dims=["field_name", "sky_dir_label"],
            )
            combined_ephemeris_field_and_source_xds["FIELD_OFFSET"].attrs = (
                combined_ephemeris_field_and_source_xds["FIELD_PHASE_CENTER"].attrs
            )
            combined_ephemeris_field_and_source_xds["FIELD_OFFSET"].attrs["units"] = (
                combined_ephemeris_field_and_source_xds["FIELD_OFFSET"].attrs["units"][
                    :2
                ]
            )

            ra1 = (
                combined_ephemeris_field_and_source_xds["FIELD_OFFSET"]
                .sel(sky_dir_label="ra")
                .values
            )
            dec1 = (
                combined_ephemeris_field_and_source_xds["FIELD_OFFSET"]
                .sel(sky_dir_label="dec")
                .values
            )
            ra2 = 0.0
            dec2 = 0.0

            from xradio._utils.coord_math import haversine

            distance = haversine(ra1, dec1, ra2, dec2)
            min_index = distance.argmin()

            combined_ephemeris_field_and_source_xds.attrs["center_field_name"] = (
                combined_ephemeris_field_and_source_xds.field_name[min_index].values
            )

        return combined_field_and_source_xds, combined_ephemeris_field_and_source_xds

    def plot_phase_centers(self, label_all_fields=False, data_group="base"):
        """
        Plot the phase center locations of all fields in the Processing Set.

        This method is primarily used for visualizing mosaics. It generates scatter plots of
        the phase center coordinates for both standard and ephemeris fields. The central field
        is highlighted in red based on the closest phase center calculation.

        Parameters
        ----------
        label_all_fields : bool, optional
            If `True`, all fields will be labeled on the plot. Default is `False`.
        data_group : str, optional
            The data group to use for processing. Default is "base".

        Returns
        -------
        None

        Raises
        ------
        ValueError
            If the combined datasets are empty or improperly formatted.
        """
        combined_field_and_source_xds, combined_ephemeris_field_and_source_xds = (
            self.get_combined_field_and_source_xds(data_group)
        )
        from matplotlib import pyplot as plt

        if (len(combined_field_and_source_xds.data_vars) > 0) and (
            "FIELD_PHASE_CENTER" in combined_field_and_source_xds
        ):
            plt.figure()
            plt.title("Field Phase Center Locations")
            plt.scatter(
                combined_field_and_source_xds["FIELD_PHASE_CENTER"].sel(
                    sky_dir_label="ra"
                ),
                combined_field_and_source_xds["FIELD_PHASE_CENTER"].sel(
                    sky_dir_label="dec"
                ),
            )

            center_field_name = combined_field_and_source_xds.attrs["center_field_name"]
            center_field = combined_field_and_source_xds.sel(
                field_name=center_field_name
            )
            plt.scatter(
                center_field["FIELD_PHASE_CENTER"].sel(sky_dir_label="ra"),
                center_field["FIELD_PHASE_CENTER"].sel(sky_dir_label="dec"),
                color="red",
                label=center_field_name,
            )
            plt.xlabel("RA (rad)")
            plt.ylabel("DEC (rad)")
            plt.legend()
            plt.show()

        if (len(combined_ephemeris_field_and_source_xds.data_vars) > 0) and (
            "FIELD_PHASE_CENTER" in combined_ephemeris_field_and_source_xds
        ):

            plt.figure()
            plt.title(
                "Offset of Field Phase Center from Source Location (Ephemeris Data)"
            )
            plt.scatter(
                combined_ephemeris_field_and_source_xds["FIELD_OFFSET"].sel(
                    sky_dir_label="ra"
                ),
                combined_ephemeris_field_and_source_xds["FIELD_OFFSET"].sel(
                    sky_dir_label="dec"
                ),
            )

            center_field_name = combined_ephemeris_field_and_source_xds.attrs[
                "center_field_name"
            ]
            center_field = combined_ephemeris_field_and_source_xds.sel(
                field_name=center_field_name
            )
            plt.scatter(
                center_field["FIELD_OFFSET"].sel(sky_dir_label="ra"),
                center_field["FIELD_OFFSET"].sel(sky_dir_label="dec"),
                color="red",
                label=center_field_name,
            )
            plt.xlabel("RA Offset (rad)")
            plt.ylabel("DEC Offset (rad)")
            plt.legend()
            plt.show()

    def get_combined_antenna_xds(self):
        """
        Combine the `antenna_xds` datasets from all Measurement Sets into a single dataset.

        This method concatenates the antenna datasets from each Measurement Set along the 'antenna_name' dimension.

        Returns
        -------
        xarray.Dataset
            A combined `xarray.Dataset` containing antenna information from all Measurement Sets.

        Raises
        ------
        ValueError
            If antenna datasets are missing required variables or improperly formatted.
        """
        combined_antenna_xds = xr.Dataset()
        for cor_name, ms_xds in self.items():
            antenna_xds = ms_xds.antenna_xds.copy(deep=True)

            if len(combined_antenna_xds.data_vars) == 0:
                combined_antenna_xds = antenna_xds
            else:
                combined_antenna_xds = xr.concat(
                    [combined_antenna_xds, antenna_xds],
                    dim="antenna_name",
                    data_vars="minimal",
                    coords="minimal",
                )

        # ALMA WVR antenna_xds data has a NaN value for the antenna receptor angle.
        if "ANTENNA_RECEPTOR_ANGLE" in combined_antenna_xds.data_vars:
            combined_antenna_xds = combined_antenna_xds.dropna("antenna_name")

        combined_antenna_xds = combined_antenna_xds.drop_duplicates("antenna_name")

        return combined_antenna_xds

    def plot_antenna_positions(self):
        """
        Plot the antenna positions of all antennas in the Processing Set.

        This method generates three scatter plots displaying the antenna positions in different planes:
        - X vs Y
        - X vs Z
        - Y vs Z

        Parameters
        ----------
        None

        Returns
        -------
        None

        Raises
        ------
        ValueError
            If the combined antenna dataset is empty or missing required coordinates.
        """
        combined_antenna_xds = self.get_combined_antenna_xds()
        from matplotlib import pyplot as plt

        plt.figure()
        plt.title("Antenna Positions")
        plt.scatter(
            combined_antenna_xds["ANTENNA_POSITION"].sel(cartesian_pos_label="x"),
            combined_antenna_xds["ANTENNA_POSITION"].sel(cartesian_pos_label="y"),
        )
        plt.xlabel("x (m)")
        plt.ylabel("y (m)")
        plt.show()

        plt.figure()
        plt.title("Antenna Positions")
        plt.scatter(
            combined_antenna_xds["ANTENNA_POSITION"].sel(cartesian_pos_label="x"),
            combined_antenna_xds["ANTENNA_POSITION"].sel(cartesian_pos_label="z"),
        )
        plt.xlabel("x (m)")
        plt.ylabel("z (m)")
        plt.show()

        plt.figure()
        plt.title("Antenna Positions")
        plt.scatter(
            combined_antenna_xds["ANTENNA_POSITION"].sel(cartesian_pos_label="y"),
            combined_antenna_xds["ANTENNA_POSITION"].sel(cartesian_pos_label="z"),
        )
        plt.xlabel("y (m)")
        plt.ylabel("z (m)")
        plt.show()



"""Typing support for xarray data classes

This has been extracted from the xarray-dataclasses package by astropenguin
(see https://github.com/astropenguin/xarray-dataclasses/). The reason we
replicate this here is because we actually ignore / redo everything but the
type annotations, especially adding xradio-specific support for multiple
options in data variable / coordinate dimensionality and dtype.
"""

from typing import (
    Any,
    List,
    Tuple,
    Hashable,
    Iterable,
    Type,
    ClassVar,
    Dict,
    TypeVar,
    Union,
    Sequence,
    Generic,
    Collection,
    Literal,
    get_type_hints,
    get_args,
    get_origin,
    Annotated,
    Protocol,
)

from typing import Union

try:
    # Python 3.10 forward: TypeAlias, ParamSpec are standard, and there is the
    # "a | b" UnionType alternative to "Union[a,b]"
    from typing import TypeAlias, ParamSpec
    from types import UnionType

    HAVE_UNIONTYPE = True
except ImportError:
    # Python 3.9: Get TypeAlias, ParamSpec from typing_extensions, no support
    # for "a | b"
    from typing_extensions import (
        TypeAlias,
        ParamSpec,
    )

    HAVE_UNIONTYPE = False
import numpy as np
from itertools import chain
from enum import Enum

PInit = ParamSpec("PInit")
T = TypeVar("T")
TDataClass = TypeVar("TDataClass", bound="DataClass[Any]")
TDims = TypeVar("TDims", covariant=True)
TDType = TypeVar("TDType", covariant=True)
THashable = TypeVar("THashable", bound=Hashable)

AnyArray: TypeAlias = "np.ndarray[Any, Any]"
AnyDType: TypeAlias = "np.dtype[Any]"
AnyField: TypeAlias = "Field[Any]"
AnyXarray: TypeAlias = "xr.DataArray | xr.Dataset"
Dims = Tuple[str, ...]
Order = Literal["C", "F"]
Shape = Union[Sequence[int], int]
Sizes = Dict[str, int]


class DataClass(Protocol[PInit]):
    """Type hint for dataclass objects."""

    def __init__(self, *args: PInit.args, **kwargs: PInit.kwargs) -> None: ...

    __dataclass_fields__: ClassVar[Dict[str, AnyField]]


class Labeled(Generic[TDims]):
    """Type hint for labeled objects."""

    pass


# type hints (public)
class Role(Enum):
    """Annotations for typing dataclass fields."""

    ATTR = "attr"
    """Annotation for attribute fields."""

    COORD = "coord"
    """Annotation for coordinate fields."""

    DATA = "data"
    """Annotation for data (variable) fields."""

    NAME = "name"
    """Annotation for name fields."""

    OTHER = "other"
    """Annotation for other fields."""

    @classmethod
    def annotates(cls, tp: Any) -> bool:
        """Check if any role annotates a type hint."""
        if get_origin(tp) is not Annotated:
            return False

        return any(isinstance(arg, cls) for arg in get_args(tp))


Attr = Annotated[T, Role.ATTR]
"""Type hint for attribute fields (``Attr[T]``).

Example:
    ::

        @dataclass
        class Image(AsDataArray):
            data: Data[tuple[X, Y], float]
            long_name: Attr[str] = "luminance"
            units: Attr[str] = "cd / m^2"

Hint:
    The following field names are specially treated when plotting.

    - ``long_name`` or ``standard_name``: Coordinate name.
    - ``units``: Coordinate units.

Reference:
    https://xarray.pydata.org/en/stable/user-guide/plotting.html

"""

Coord = Annotated[Union[Labeled[TDims], Collection[TDType], TDType], Role.COORD]
"""Type hint for coordinate fields (``Coord[TDims, TDType]``).

Example:
    ::

        @dataclass
        class Image(AsDataArray):
            data: Data[tuple[X, Y], float]
            mask: Coord[tuple[X, Y], bool]
            x: Coord[X, int] = 0
            y: Coord[Y, int] = 0

Hint:
    A coordinate field whose name is the same as ``TDims``
    (e.g. ``x: Coord[X, int]``) can define a dimension.

"""

Coordof = Annotated[Union[TDataClass, Any], Role.COORD]
"""Type hint for coordinate fields (``Coordof[TDataClass]``).

Unlike ``Coord``, it specifies a dataclass that defines a DataArray class.
This is useful when users want to add metadata to dimensions for plotting.

Example:
    ::

        @dataclass
        class XAxis:
            data: Data[X, int]
            long_name: Attr[str] = "x axis"


        @dataclass
        class YAxis:
            data: Data[Y, int]
            long_name: Attr[str] = "y axis"


        @dataclass
        class Image(AsDataArray):
            data: Data[tuple[X, Y], float]
            x: Coordof[XAxis] = 0
            y: Coordof[YAxis] = 0

Hint:
    A class used in ``Coordof`` does not need to inherit ``AsDataArray``.

"""

Data = Annotated[Union[Labeled[TDims], Collection[TDType], TDType], Role.DATA]
"""Type hint for data fields (``Coordof[TDims, TDType]``).

Example:
    Exactly one data field is allowed in a DataArray class
    (the second and subsequent data fields are just ignored)::

        @dataclass
        class Image(AsDataArray):
            data: Data[tuple[X, Y], float]

    Multiple data fields are allowed in a Dataset class::

        @dataclass
        class ColorImage(AsDataset):
            red: Data[tuple[X, Y], float]
            green: Data[tuple[X, Y], float]
            blue: Data[tuple[X, Y], float]

"""

Dataof = Annotated[Union[TDataClass, Any], Role.DATA]
"""Type hint for data fields (``Coordof[TDataClass]``).

Unlike ``Data``, it specifies a dataclass that defines a DataArray class.
This is useful when users want to reuse a dataclass in a Dataset class.

Example:
    ::

        @dataclass
        class Image:
            data: Data[tuple[X, Y], float]
            x: Coord[X, int] = 0
            y: Coord[Y, int] = 0


        @dataclass
        class ColorImage(AsDataset):
            red: Dataof[Image]
            green: Dataof[Image]
            blue: Dataof[Image]

Hint:
    A class used in ``Dataof`` does not need to inherit ``AsDataArray``.

"""

Name = Annotated[THashable, Role.NAME]
"""Type hint for name fields (``Name[THashable]``).

Example:
    ::

        @dataclass
        class Image(AsDataArray):
            data: Data[tuple[X, Y], float]
            name: Name[str] = "image"

"""


def deannotate(tp: Any) -> Any:
    """Recursively remove annotations in a type hint."""

    class Temporary:
        __annotations__ = dict(type=tp)

    return get_type_hints(Temporary)["type"]


def find_annotated(tp: Any) -> Iterable[Any]:
    """Generate all annotated types in a type hint."""
    args = get_args(tp)

    if get_origin(tp) is Annotated:
        yield tp
        yield from find_annotated(args[0])
    else:
        yield from chain(*map(find_annotated, args))


def get_annotated(tp: Any) -> Any:
    """Extract the first role-annotated type."""
    for annotated in filter(Role.annotates, find_annotated(tp)):
        return deannotate(annotated)

    raise TypeError("Could not find any role-annotated type.")


def get_annotations(tp: Any) -> Tuple[Any, ...]:
    """Extract annotations of the first role-annotated type."""
    for annotated in filter(Role.annotates, find_annotated(tp)):
        return get_args(annotated)[1:]

    raise TypeError("Could not find any role-annotated type.")


def get_dataclass(tp: Any) -> Type[DataClass[Any]]:
    """Extract a dataclass."""
    try:
        dataclass = get_args(get_annotated(tp))[0]
    except TypeError:
        raise TypeError(f"Could not find any dataclass in {tp!r}.")

    if not is_dataclass(dataclass):
        raise TypeError(f"Could not find any dataclass in {tp!r}.")

    return dataclass


def get_dims(tp: Any) -> List[Dims]:
    """Extract data dimensions (dims)."""
    try:
        dims = get_args(get_args(get_annotated(tp))[0])[0]
    except TypeError:
        raise TypeError(f"Could not find any dims in {tp!r}.")

    # List of allowed dtypes (might just be one)
    if get_origin(dims) is Union:
        dims_in = get_args(dims)
    elif HAVE_UNIONTYPE and get_origin(dims) is UnionType:
        dims_in = get_args(dims)
    else:
        dims_in = [dims]

    dims_out = []
    for dim in dims_in:
        args = get_args(dim)
        origin = get_origin(dim)

        # One-dimensional dimension
        if origin is Literal:
            dims_out.append((str(args[0]),))
            continue

        if not (origin is tuple or origin is Tuple):
            raise TypeError(f"Could not find any dims in {tp!r}.")

        # Zero-dimensions
        if args == () or args == ((),):
            dims_out.append(())
            continue

        if not all(get_origin(arg) is Literal for arg in args):
            raise TypeError(f"Could not find any dims in {tp!r}.")

        dims_out.append(tuple(str(get_args(arg)[0]) for arg in args))

    return dims_out


def get_types(tp: Any) -> List[AnyDType]:
    """Extract data types from type annotation

    E.g. Coord[..., Type1 | Type2 | ...] or Data[..., Type1 | Type2 | ...]

    """
    try:
        typ = get_args(get_args(get_annotated(tp))[1])[0]
    except TypeError:
        raise TypeError(f"Could not find any dtype in {tp!r}.")

    # List of allowed dtypes (might just be one)
    if get_origin(typ) is Union:
        types_in = get_args(typ)
    elif HAVE_UNIONTYPE and get_origin(typ) is UnionType:
        types_in = get_args(typ)
    else:
        types_in = [typ]

    types_out = []
    for dt in types_in:
        # Handle case that we want to allow "Any"
        if dt is Any or dt is type(None):
            types_out.append(None)
            continue

        # Allow specifying type as literal (e.g. string)
        elif get_origin(dt) is Literal:
            dt = get_args(dt)[0]

        # Return type
        types_out.append(dt)

    return types_out


def get_name(tp: Any, default: Hashable = None) -> Hashable:
    """Extract a name if found or return given default."""
    try:
        annotations = get_annotations(tp)[1:]
    except TypeError:
        return default

    for annotation in annotations:
        if isinstance(annotation, Hashable):
            return annotation

    return default


def get_role(tp: Any, default: Role = Role.OTHER) -> Role:
    """Extract a role if found or return given default."""
    try:
        return get_annotations(tp)[0]
    except TypeError:
        return default


def is_optional(type_ann):
    """
    Check whether a type annotation indicates that the value is optional

    Boils down to checking whether it's a union type that includes None
    """
    if get_origin(type_ann) is Union:
        return None.__class__ in get_args(type_ann)
    return False
